#!/usr/bin/env python3
#
# Copyright (c) 2011-2014 David Bremner <david@tethera.net>
#                         W. Trevor King <wking@tremily.us>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see https://www.gnu.org/licenses/ .

"""
Manage notmuch tags with Git
"""

from __future__ import print_function
from __future__ import unicode_literals

import codecs as _codecs
import collections as _collections
import functools as _functools
import inspect as _inspect
import locale as _locale
import logging as _logging
import os as _os
import re as _re
import shutil as _shutil
import subprocess as _subprocess
import sys as _sys
import tempfile as _tempfile
import textwrap as _textwrap
from urllib.parse import quote as _quote
from urllib.parse import unquote as _unquote
import json as _json

_LOG = _logging.getLogger('notmuch-git')
_LOG.setLevel(_logging.WARNING)
_LOG.addHandler(_logging.StreamHandler())

NOTMUCH_GIT_DIR = None
TAG_PREFIX = None
FORMAT_VERSION = 1

_HEX_ESCAPE_REGEX = _re.compile('%[0-9A-F]{2}')
_TAG_DIRECTORY = 'tags/'
_TAG_FILE_REGEX = ( _re.compile(_TAG_DIRECTORY + '(?P<id>[^/]*)/(?P<tag>[^/]*)'),
                    _re.compile(_TAG_DIRECTORY + '([0-9a-f]{2}/){2}(?P<id>[^/]*)/(?P<tag>[^/]*)'))

# magic hash for Git (git hash-object -t blob /dev/null)
_EMPTYBLOB = 'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391'

def _hex_quote(string, safe='+@=:,'):
    """
    quote('abc def') -> 'abc%20def'.

    Wrap urllib.parse.quote with additional safe characters (in
    addition to letters, digits, and '_.-') and lowercase hex digits
    (e.g. '%3a' instead of '%3A').
    """
    uppercase_escapes = _quote(string, safe)
    return _HEX_ESCAPE_REGEX.sub(
        lambda match: match.group(0).lower(),
        uppercase_escapes)

def _xapian_quote(string):
    """
    Quote a string for Xapian's QueryParser.

    Xapian uses double-quotes for quoting strings.  You can escape
    internal quotes by repeating them [1,2,3].

    [1]: https://trac.xapian.org/ticket/128#comment:2
    [2]: https://trac.xapian.org/ticket/128#comment:17
    [3]: https://trac.xapian.org/changeset/13823/svn
    """
    return '"{0}"'.format(string.replace('"', '""'))


def _xapian_unquote(string):
    """
    Unquote a Xapian-quoted string.
    """
    if string.startswith('"') and string.endswith('"'):
        return string[1:-1].replace('""', '"')
    return string


def timed(fn):
    """Timer decorator"""
    from time import perf_counter

    def inner(*args, **kwargs):
        start_time = perf_counter()
        rval = fn(*args, **kwargs)
        end_time = perf_counter()
        _LOG.info('{0}: {1:.8f}s elapsed'.format(fn.__name__, end_time - start_time))
        return rval

    return inner


class SubprocessError(RuntimeError):
    "A subprocess exited with a nonzero status"
    def __init__(self, args, status, stdout=None, stderr=None):
        self.status = status
        self.stdout = stdout
        self.stderr = stderr
        msg = '{args} exited with {status}'.format(args=args, status=status)
        if stderr:
            msg = '{msg}: {stderr}'.format(msg=msg, stderr=stderr)
        super(SubprocessError, self).__init__(msg)


class _SubprocessContextManager(object):
    """
    PEP 343 context manager for subprocesses.

    'expect' holds a tuple of acceptable exit codes, otherwise we'll
    raise a SubprocessError in __exit__.
    """
    def __init__(self, process, args, expect=(0,)):
        self._process = process
        self._args = args
        self._expect = expect

    def __enter__(self):
        return self._process

    def __exit__(self, type, value, traceback):
        for name in ['stdin', 'stdout', 'stderr']:
            stream = getattr(self._process, name)
            if stream:
                stream.close()
                setattr(self._process, name, None)
        status = self._process.wait()
        _LOG.debug(
            'collect {args} with status {status} (expected {expect})'.format(
                args=self._args, status=status, expect=self._expect))
        if status not in self._expect:
            raise SubprocessError(args=self._args, status=status)

    def wait(self):
        return self._process.wait()


def _spawn(args, input=None, additional_env=None, wait=False, stdin=None,
           stdout=None, stderr=None, encoding=_locale.getpreferredencoding(),
           expect=(0,), **kwargs):
    """Spawn a subprocess, and optionally wait for it to finish.

    This wrapper around subprocess.Popen has two modes, depending on
    the truthiness of 'wait'.  If 'wait' is true, we use p.communicate
    internally to write 'input' to the subprocess's stdin and read
    from it's stdout/stderr.  If 'wait' is False, we return a
    _SubprocessContextManager instance for fancier handling
    (e.g. piping between processes).

    For 'wait' calls when you want to write to the subprocess's stdin,
    you only need to set 'input' to your content.  When 'input' is not
    None but 'stdin' is, we'll automatically set 'stdin' to PIPE
    before calling Popen.  This avoids having the subprocess
    accidentally inherit the launching process's stdin.
    """
    _LOG.debug('spawn {args} (additional env. var.: {env})'.format(
        args=args, env=additional_env))
    if not stdin and input is not None:
        stdin = _subprocess.PIPE
    if additional_env:
        if not kwargs.get('env'):
            kwargs['env'] = dict(_os.environ)
        kwargs['env'].update(additional_env)
    p = _subprocess.Popen(
        args, stdin=stdin, stdout=stdout, stderr=stderr, **kwargs)
    if wait:
        if hasattr(input, 'encode'):
            input = input.encode(encoding)
        (stdout, stderr) = p.communicate(input=input)
        status = p.wait()
        _LOG.debug(
            'collect {args} with status {status} (expected {expect})'.format(
                args=args, status=status, expect=expect))
        if stdout is not None:
            stdout = stdout.decode(encoding)
        if stderr is not None:
            stderr = stderr.decode(encoding)
        if status not in expect:
            raise SubprocessError(
                args=args, status=status, stdout=stdout, stderr=stderr)
        return (status, stdout, stderr)
    if p.stdin and not stdin:
        p.stdin.close()
        p.stdin = None
    if p.stdin:
        p.stdin = _codecs.getwriter(encoding=encoding)(stream=p.stdin)
    stream_reader = _codecs.getreader(encoding=encoding)
    if p.stdout:
        p.stdout = stream_reader(stream=p.stdout)
    if p.stderr:
        p.stderr = stream_reader(stream=p.stderr)
    return _SubprocessContextManager(args=args, process=p, expect=expect)


def _git(args, **kwargs):
    args = ['git', '--git-dir', NOTMUCH_GIT_DIR] + list(args)
    return _spawn(args=args, **kwargs)


def _get_current_branch():
    """Get the name of the current branch.

    Return 'None' if we're not on a branch.
    """
    try:
        (status, branch, stderr) = _git(
            args=['symbolic-ref', '--short', 'HEAD'],
            stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, wait=True)
    except SubprocessError as e:
        if 'not a symbolic ref' in e:
            return None
        raise
    return branch.strip()


def _get_remote():
    "Get the default remote for the current branch."
    local_branch = _get_current_branch()
    (status, remote, stderr) = _git(
        args=['config', 'branch.{0}.remote'.format(local_branch)],
        stdout=_subprocess.PIPE, wait=True)
    return remote.strip()

def _tag_query(prefix=None):
    if prefix is None:
        prefix = TAG_PREFIX
    return '(tag (starts-with "{:s}"))'.format(prefix.replace('"','\\\"'))

def count_messages(prefix=None):
    "count messages with a given prefix."
    (status, stdout, stderr) = _spawn(
        args=['notmuch', 'count', '--query=sexp', _tag_query(prefix)],
        stdout=_subprocess.PIPE, wait=True)
    if status != 0:
        _LOG.error("failed to run notmuch config")
        sys.exit(1)
    return int(stdout.rstrip())

def get_tags(prefix=None):
    "Get a list of tags with a given prefix."
    (status, stdout, stderr) = _spawn(
        args=['notmuch', 'search', '--exclude=false', '--query=sexp', '--output=tags', _tag_query(prefix)],
        stdout=_subprocess.PIPE, wait=True)
    return [tag for tag in stdout.splitlines()]

def archive(treeish='HEAD', args=()):
    """
    Dump a tar archive of the current notmuch-git tag set.

    Using 'git archive'.

    Each tag $tag for message with Message-Id $id is written to
    an empty file

      tags/hash1(id)/hash2(id)/encode($id)/encode($tag)

    The encoding preserves alphanumerics, and the characters
    "+-_@=.:," (not the quotes).  All other octets are replaced with
    '%' followed by a two digit hex number.
    """
    _git(args=['archive', treeish] + list(args), wait=True)


def clone(repository):
    """
    Create a local notmuch-git repository from a remote source.

    This wraps 'git clone', adding some options to avoid creating a
    working tree while preserving remote-tracking branches and
    upstreams.
    """
    with _tempfile.TemporaryDirectory(prefix='notmuch-git-clone.') as workdir:
        _spawn(
            args=[
                'git', 'clone', '--no-checkout', '--separate-git-dir', NOTMUCH_GIT_DIR,
                repository, workdir],
            wait=True)
    _git(args=['config', '--unset', 'core.worktree'], wait=True, expect=(0, 5))
    _git(args=['config', 'core.bare', 'true'], wait=True)
    (status, stdout, stderr) = _git(args=['show-ref', '--verify',
                                          '--quiet',
                                          'refs/remotes/origin/config'],
                                    expect=(0,1),
                                    wait=True)
    if status == 0:
        _git(args=['branch', 'config', 'origin/config'], wait=True)
    existing_tags = get_tags()
    if existing_tags:
        _LOG.warning(
            'Not checking out to avoid clobbering existing tags: {}'.format(
            ', '.join(existing_tags)))
    else:
        checkout()


def _is_committed(status):
    return len(status['added']) + len(status['deleted']) == 0


class CachedIndex:
    def __init__(self, repo, treeish):
        self.cache_path = _os.path.join(repo, 'notmuch', 'index_cache.json')
        self.index_path = _os.path.join(repo, 'index')
        self.current_treeish = treeish
        # cached values
        self.treeish = None
        self.hash = None
        self.index_checksum = None

        self._load_cache_file()

    def _load_cache_file(self):
        try:
            with open(self.cache_path) as f:
                data = _json.load(f)
                self.treeish = data['treeish']
                self.hash = data['hash']
                self.index_checksum = data['index_checksum']
        except FileNotFoundError:
            pass
        except _json.JSONDecodeError:
            _LOG.error("Error decoding cache")
            _sys.exit(1)

    def __enter__(self):
        self.read_tree()
        return self

    def __exit__(self, type, value, traceback):
        checksum = _read_index_checksum(self.index_path)
        (_, hash, _) = _git(
            args=['rev-parse', self.current_treeish],
            stdout=_subprocess.PIPE,
            wait=True)

        with open(self.cache_path, "w") as f:
            _json.dump({'treeish': self.current_treeish,
                        'hash': hash.rstrip(),  'index_checksum': checksum }, f)

    @timed
    def read_tree(self):
        current_checksum = _read_index_checksum(self.index_path)
        (_, hash, _) = _git(
            args=['rev-parse', self.current_treeish],
            stdout=_subprocess.PIPE,
            wait=True)
        current_hash = hash.rstrip()

        if self.current_treeish == self.treeish and \
           self.index_checksum and self.index_checksum == current_checksum and \
           self.hash and self.hash == current_hash:
            return

        _git(args=['read-tree', self.current_treeish], wait=True)


def check_safe_fraction(status):
    safe = 0.1
    conf = _notmuch_config_get ('git.safe_fraction')
    if conf and conf != '':
        safe=float(conf)

    total = count_messages (TAG_PREFIX)
    if total == 0:
        _LOG.error('No existing tags with given prefix, stopping.'.format(safe))
        _LOG.error('Use --force to override.')
        exit(1)
    change = len(status['added'])+len(status['deleted'])
    fraction = change/total
    _LOG.debug('total messages {:d}, change: {:d}, fraction: {:f}'.format(total,change,fraction))
    if fraction > safe:
        _LOG.error('safe fraction {:f} exceeded, stopping.'.format(safe))
        _LOG.error('Use --force to override or reconfigure git.safe_fraction.')
        exit(1)

def commit(treeish='HEAD', message=None, force=False):
    """
    Commit prefix-matching tags from the notmuch database to Git.
    """

    status = get_status()

    if _is_committed(status=status):
        _LOG.warning('Nothing to commit')
        return

    if not force:
        check_safe_fraction (status)

    with CachedIndex(NOTMUCH_GIT_DIR, treeish) as index:
        try:
            _update_index(status=status)
            (_, tree, _) = _git(
                args=['write-tree'],
                stdout=_subprocess.PIPE,
                wait=True)
            (_, parent, _) = _git(
                args=['rev-parse', treeish],
                stdout=_subprocess.PIPE,
                wait=True)
            (_, commit, _) = _git(
                args=['commit-tree', tree.strip(), '-p', parent.strip()],
                input=message,
                stdout=_subprocess.PIPE,
                wait=True)
            _git(
                args=['update-ref', treeish, commit.strip()],
                stdout=_subprocess.PIPE,
                wait=True)
        except Exception as e:
            _git(args=['read-tree', '--empty'], wait=True)
            _git(args=['read-tree', treeish], wait=True)
            raise

@timed
def _update_index(status):
    with _git(
            args=['update-index', '--index-info'],
            stdin=_subprocess.PIPE) as p:
        for id, tags in status['deleted'].items():
            for line in _index_tags_for_message(id=id, status='D', tags=tags):
                p.stdin.write(line)
        for id, tags in status['added'].items():
            for line in _index_tags_for_message(id=id, status='A', tags=tags):
                p.stdin.write(line)


def fetch(remote=None):
    """
    Fetch changes from the remote repository.

    See 'merge' to bring those changes into notmuch.
    """
    args = ['fetch']
    if remote:
        args.append(remote)
    _git(args=args, wait=True)


def init(remote=None,format_version=None):
    """
    Create an empty notmuch-git repository.

    This wraps 'git init' with a few extra steps to support subsequent
    status and commit commands.
    """
    from pathlib import Path
    parent = Path(NOTMUCH_GIT_DIR).parent
    try:
        _os.makedirs(parent)
    except FileExistsError:
        pass

    if not format_version:
        format_version = 1

    format_version=int(format_version)

    if format_version > 1 or format_version < 0:
        _LOG.error("Illegal format version {:d}".format(format_version))
        _sys.exit(1)

    _spawn(args=['git', '--git-dir', NOTMUCH_GIT_DIR, 'init',
                 '--initial-branch=master', '--quiet', '--bare'], wait=True)
    _git(args=['config', 'core.logallrefupdates', 'true'], wait=True)
    # create an empty blob (e69de29bb2d1d6434b8b29ae775ad8c2e48c5391)
    _git(args=['hash-object', '-w', '--stdin'], input='', wait=True)
    allow_empty=('--allow-empty',)
    if format_version >= 1:
        allow_empty=()
        # create a blob for the FORMAT file
        (status, stdout, _) = _git(args=['hash-object', '-w', '--stdin'], stdout=_subprocess.PIPE,
                                   input='{:d}\n'.format(format_version), wait=True)
        verhash=stdout.rstrip()
        _LOG.debug('hash of FORMAT blob = {:s}'.format(verhash))
        # Add FORMAT to the index
        _git(args=['update-index', '--add', '--cacheinfo', '100644,{:s},FORMAT'.format(verhash)], wait=True)

    _git(
        args=[
            'commit', *allow_empty, '-m', 'Start a new notmuch-git repository'
        ],
        additional_env={'GIT_WORK_TREE': NOTMUCH_GIT_DIR},
        wait=True)


def checkout(force=None):
    """
    Update the notmuch database from Git.

    This is mainly useful to discard your changes in notmuch relative
    to Git.
    """
    status = get_status()

    if not force:
        check_safe_fraction(status)

    with _spawn(
            args=['notmuch', 'tag', '--batch'], stdin=_subprocess.PIPE) as p:
        for id, tags in status['added'].items():
            p.stdin.write(_batch_line(action='-', id=id, tags=tags))
        for id, tags in status['deleted'].items():
            p.stdin.write(_batch_line(action='+', id=id, tags=tags))


def _batch_line(action, id, tags):
    """
    'notmuch tag --batch' line for adding/removing tags.

    Set 'action' to '-' to remove a tag or '+' to add the tags to a
    given message id.
    """
    tag_string = ' '.join(
        '{action}{prefix}{tag}'.format(
            action=action, prefix=_ENCODED_TAG_PREFIX, tag=_hex_quote(tag))
        for tag in tags)
    line = '{tags} -- id:{id}\n'.format(
        tags=tag_string, id=_xapian_quote(string=id))
    return line


def _insist_committed():
    "Die if the the notmuch tags don't match the current HEAD."
    status = get_status()
    if not _is_committed(status=status):
        _LOG.error('\n'.join([
            'Uncommitted changes to {prefix}* tags in notmuch',
            '',
            "For a summary of changes, run 'notmuch-git status'",
            "To save your changes,     run 'notmuch-git commit' before merging/pull",
            "To discard your changes,  run 'notmuch-git checkout'",
            ]).format(prefix=TAG_PREFIX))
        _sys.exit(1)


def pull(repository=None, refspecs=None):
    """
    Pull (merge) remote repository changes to notmuch.

    'pull' is equivalent to 'fetch' followed by 'merge'.  We use the
    Git-configured repository for your current branch
    (branch.<name>.repository, likely 'origin', and
    branch.<name>.merge, likely 'master').
    """
    _insist_committed()
    if refspecs and not repository:
        repository = _get_remote()
    args = ['pull']
    if repository:
        args.append(repository)
    if refspecs:
        args.extend(refspecs)
    with _tempfile.TemporaryDirectory(prefix='notmuch-git-pull.') as workdir:
        for command in [
                ['reset', '--hard'],
                args]:
            _git(
                args=command,
                additional_env={'GIT_WORK_TREE': workdir},
                wait=True)
    checkout()


def merge(reference='@{upstream}'):
    """
    Merge changes from 'reference' into HEAD and load the result into notmuch.

    The default reference is '@{upstream}'.
    """
    _insist_committed()
    with _tempfile.TemporaryDirectory(prefix='notmuch-git-merge.') as workdir:
        for command in [
                ['reset', '--hard'],
                ['merge', reference]]:
            _git(
                args=command,
                additional_env={'GIT_WORK_TREE': workdir},
                wait=True)
    checkout()


def log(args=()):
    """
    A simple wrapper for 'git log'.

    After running 'notmuch-git fetch', you can inspect the changes with
    'notmuch-git log HEAD..@{upstream}'.
    """
    # we don't want output trapping here, because we want the pager.
    args = ['log', '--name-status', '--no-renames'] + list(args)
    with _git(args=args, expect=(0, 1, -13)) as p:
        p.wait()


def push(repository=None, refspecs=None):
    "Push the local notmuch-git Git state to a remote repository."
    if refspecs and not repository:
        repository = _get_remote()
    args = ['push']
    if repository:
        args.append(repository)
    if refspecs:
        args.extend(refspecs)
    _git(args=args, wait=True)


def status():
    """
    Show pending updates in notmuch or git repo.

    Prints lines of the form

      ng Message-Id tag

    where n is a single character representing notmuch database status

    * A

      Tag is present in notmuch database, but not committed to notmuch-git
      (equivalently, tag has been deleted in notmuch-git repo, e.g. by a
      pull, but not restored to notmuch database).

    * D

      Tag is present in notmuch-git repo, but not restored to notmuch
      database (equivalently, tag has been deleted in notmuch).

    * U

      Message is unknown (missing from local notmuch database).

    The second character (if present) represents a difference between
    local and upstream branches. Typically 'notmuch-git fetch' needs to be
    run to update this.

    * a

      Tag is present in upstream, but not in the local Git branch.

    * d

      Tag is present in local Git branch, but not upstream.
    """
    status = get_status()
    # 'output' is a nested defaultdict for message status:
    # * The outer dict is keyed by message id.
    # * The inner dict is keyed by tag name.
    # * The inner dict values are status strings (' a', 'Dd', ...).
    output = _collections.defaultdict(
        lambda : _collections.defaultdict(lambda : ' '))
    for id, tags in status['added'].items():
        for tag in tags:
            output[id][tag] = 'A'
    for id, tags in status['deleted'].items():
        for tag in tags:
            output[id][tag] = 'D'
    for id, tags in status['missing'].items():
        for tag in tags:
            output[id][tag] = 'U'
    if _is_unmerged():
        for id, tag in _diff_refs(filter='A'):
            output[id][tag] += 'a'
        for id, tag in _diff_refs(filter='D'):
            output[id][tag] += 'd'
    for id, tag_status in sorted(output.items()):
        for tag, status in sorted(tag_status.items()):
            print('{status}\t{id}\t{tag}'.format(
                status=status, id=id, tag=tag))


def _is_unmerged(ref='@{upstream}'):
    try:
        (status, fetch_head, stderr) = _git(
            args=['rev-parse', ref],
            stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, wait=True)
    except SubprocessError as e:
        if 'No upstream configured' in e.stderr:
            return
        raise
    (status, base, stderr) = _git(
        args=['merge-base', 'HEAD', ref],
        stdout=_subprocess.PIPE, wait=True)
    return base != fetch_head

class DatabaseCache:
    def __init__(self):
        try:
            from notmuch2 import Database
            self._notmuch = Database()
        except ImportError:
            self._notmuch = None
        self._known = {}

    def known(self,id):
        if id in self._known:
            return self._known[id];

        if self._notmuch:
            try:
                _ = self._notmuch.find(id)
                self._known[id] = True
            except LookupError:
                self._known[id] = False
        else:
            (_, stdout, stderr) = _spawn(
                args=['notmuch', 'search', '--exclude=false', '--output=files', 'id:{0}'.format(id)],
                stdout=_subprocess.PIPE,
                wait=True)
            self._known[id] = stdout != None
        return self._known[id]

@timed
def get_status():
    status = {
        'deleted': {},
        'missing': {},
        }
    db = DatabaseCache()
    with PrivateIndex(repo=NOTMUCH_GIT_DIR, prefix=TAG_PREFIX) as index:
        maybe_deleted = index.diff(filter='D')
        for id, tags in maybe_deleted.items():
            if db.known(id):
                status['deleted'][id] = tags
            else:
                status['missing'][id] = tags
        status['added'] = index.diff(filter='A')

    return status

class PrivateIndex:
    def __init__(self, repo, prefix):
        try:
            _os.makedirs(_os.path.join(repo, 'notmuch'))
        except FileExistsError:
            pass

        file_name = 'notmuch/index'
        self.index_path = _os.path.join(repo, file_name)
        self.cache_path = _os.path.join(repo, 'notmuch', '{:s}.json'.format(_hex_quote(file_name)))

        self.current_prefix = prefix

        self.prefix = None
        self.uuid = None
        self.lastmod = None
        self.checksum = None
        self._load_cache_file()
        self.file_tree = None
        self._index_tags()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        checksum = _read_index_checksum(self.index_path)
        (count, uuid, lastmod) = _read_database_lastmod()
        with open(self.cache_path, "w") as f:
            _json.dump({'prefix': self.current_prefix, 'uuid': uuid, 'lastmod': lastmod,  'checksum': checksum }, f)

    def _load_cache_file(self):
        try:
            with open(self.cache_path) as f:
                data = _json.load(f)
                self.prefix = data['prefix']
                self.uuid = data['uuid']
                self.lastmod = data['lastmod']
                self.checksum = data['checksum']
        except FileNotFoundError:
            return None
        except _json.JSONDecodeError:
            _LOG.error("Error decoding cache")
            _sys.exit(1)

    @timed
    def _read_file_tree(self):
        self.file_tree = {}

        with _git(
                args=['ls-files', 'tags'],
                additional_env={'GIT_INDEX_FILE': self.index_path},
                stdout=_subprocess.PIPE) as git:
            for file in git.stdout:
                dir=_os.path.dirname(file)
                tag=_os.path.basename(file).rstrip()
                if dir not in self.file_tree:
                    self.file_tree[dir]=[tag]
                else:
                    self.file_tree[dir].append(tag)


    def _clear_tags_for_message(self, id):
        """
        Clear any existing index entries for message 'id'

        Neither 'id' nor the tags in 'tags' should be encoded/escaped.
        """

        if self.file_tree == None:
            self._read_file_tree()

        dir = _id_path(id)

        if dir not in self.file_tree:
            return

        for file in self.file_tree[dir]:
            line = '0 0000000000000000000000000000000000000000\t{:s}/{:s}\n'.format(dir,file)
            yield line


    @timed
    def _index_tags(self):
        "Write notmuch tags to private git index."
        prefix = '+{0}'.format(_ENCODED_TAG_PREFIX)
        current_checksum = _read_index_checksum(self.index_path)
        if (self.prefix == None or self.prefix != self.current_prefix
            or self.checksum == None or self.checksum != current_checksum):
            _git(
                args=['read-tree', '--empty'],
                additional_env={'GIT_INDEX_FILE': self.index_path}, wait=True)

        query = _tag_query()
        clear_tags = False
        (count,uuid,lastmod) = _read_database_lastmod()
        if self.prefix == self.current_prefix and self.uuid \
           and self.uuid == uuid and self.checksum == current_checksum:
            query = '(and (infix "lastmod:{:d}..")) {:s})'.format(self.lastmod+1, query)
            clear_tags = True
        with _spawn(
                args=['notmuch', 'dump', '--format=batch-tag', '--query=sexp', '--', query],
                stdout=_subprocess.PIPE) as notmuch:
            with _git(
                    args=['update-index', '--index-info'],
                    stdin=_subprocess.PIPE,
                    additional_env={'GIT_INDEX_FILE': self.index_path}) as git:
                for line in notmuch.stdout:
                    if line.strip().startswith('#'):
                        continue
                    (tags_string, id) = [_.strip() for _ in line.split(' -- id:')]
                    tags = [
                        _unquote(tag[len(prefix):])
                        for tag in tags_string.split()
                        if tag.startswith(prefix)]
                    id = _xapian_unquote(string=id)
                    if clear_tags:
                        for line in self._clear_tags_for_message(id=id):
                            git.stdin.write(line)
                    for line in _index_tags_for_message(
                            id=id, status='A', tags=tags):
                        git.stdin.write(line)

    @timed
    def diff(self, filter):
        """
        Get an {id: {tag, ...}} dict for a given filter.

        For example, use 'A' to find added tags, and 'D' to find deleted tags.
        """
        s = _collections.defaultdict(set)
        with _git(
                args=[
                    'diff-index', '--cached', '--diff-filter', filter,
                    '--name-only', 'HEAD'],
                additional_env={'GIT_INDEX_FILE': self.index_path},
                stdout=_subprocess.PIPE) as p:
            # Once we drop Python < 3.3, we can use 'yield from' here
            for id, tag in _unpack_diff_lines(stream=p.stdout):
                s[id].add(tag)
        return s

def _read_index_checksum (index_path):
    """Read the index checksum, as defined by index-format.txt in the git source
    WARNING: assumes SHA1 repo"""
    import binascii
    try:
        with open(index_path, 'rb') as f:
            size=_os.path.getsize(index_path)
            f.seek(size-20);
            return binascii.hexlify(f.read(20)).decode('ascii')
    except FileNotFoundError:
        return None

def _read_database_lastmod():
    with _spawn(
            args=['notmuch', 'count', '--lastmod', '*'],
            stdout=_subprocess.PIPE) as notmuch:
        (count,uuid,lastmod_str) = notmuch.stdout.readline().split()
        return (count,uuid,int(lastmod_str))

def _id_path(id):
    hid=_hex_quote(string=id)
    from hashlib import blake2b

    if FORMAT_VERSION==0:
        return 'tags/{hid}'.format(hid=hid)
    elif FORMAT_VERSION==1:
        idhash = blake2b(hid.encode('utf8'), digest_size=2).hexdigest()
        return 'tags/{dir1}/{dir2}/{hid}'.format(
            hid=hid,
            dir1=idhash[0:2],dir2=idhash[2:])
    else:
        _LOG.error("Unknown format version",FORMAT_VERSION)
        _sys.exit(1)

def _index_tags_for_message(id, status, tags):
    """
    Update the Git index to either create or delete an empty file.

    Neither 'id' nor the tags in 'tags' should be encoded/escaped.
    """
    mode = '100644'
    hash = _EMPTYBLOB

    if status == 'D':
        mode = '0'
        hash = '0000000000000000000000000000000000000000'

    for tag in tags:
        path = '{ipath}/{tag}'.format(ipath=_id_path(id),tag=_hex_quote(string=tag))
        yield '{mode} {hash}\t{path}\n'.format(mode=mode, hash=hash, path=path)


def _diff_refs(filter, a='HEAD', b='@{upstream}'):
    with _git(
            args=['diff', '--diff-filter', filter, '--name-only', a, b],
            stdout=_subprocess.PIPE) as p:
        # Once we drop Python < 3.3, we can use 'yield from' here
        for id, tag in _unpack_diff_lines(stream=p.stdout):
            yield id, tag


def _unpack_diff_lines(stream):
    "Iterate through (id, tag) tuples in a diff stream."
    for line in stream:
        match = _TAG_FILE_REGEX[FORMAT_VERSION].match(line.strip())
        if not match:
            message = 'non-tag line in diff: {!r}'.format(line.strip())
            if line.startswith(_TAG_DIRECTORY):
                raise ValueError(message)
            _LOG.info(message)
            continue
        id = _unquote(match.group('id'))
        tag = _unquote(match.group('tag'))
        yield (id, tag)


def _help(parser, command=None):
    """
    Show help for an notmuch-git command.

    Because some folks prefer:

      $ notmuch-git help COMMAND

    to

      $ notmuch-git COMMAND --help
    """
    if command:
        parser.parse_args([command, '--help'])
    else:
        parser.parse_args(['--help'])

def _notmuch_config_get(key):
    (status, stdout, stderr) = _spawn(
        args=['notmuch', 'config', 'get', key],
        stdout=_subprocess.PIPE, wait=True)
    if status != 0:
        _LOG.error("failed to run notmuch config")
        _sys.exit(1)
    return stdout.rstrip()

def read_format_version():
    try:
        (status, stdout, stderr) = _git(
            args=['cat-file', 'blob', 'master:FORMAT'],
            stdout=_subprocess.PIPE, stderr=_subprocess.PIPE, wait=True)
    except SubprocessError as e:
        _LOG.debug("failed to read FORMAT file from git, assuming format version 0")
        return 0

    return int(stdout)

# based on BaseDirectory.save_data_path from pyxdg (LGPL2+)
def xdg_data_path(profile):
    resource = _os.path.join('notmuch',profile,'git')
    assert not resource.startswith('/')
    _home = _os.path.expanduser('~')
    xdg_data_home = _os.environ.get('XDG_DATA_HOME') or \
        _os.path.join(_home, '.local', 'share')
    path = _os.path.join(xdg_data_home, resource)
    return path

if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=__doc__.strip(),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        '-C', '--git-dir', metavar='REPO',
        help='Git repository to operate on.')
    parser.add_argument(
        '-p', '--tag-prefix', metavar='PREFIX',
        default = None,
        help='Prefix of tags to operate on.')
    parser.add_argument(
        '-N', '--nmbug', action='store_true',
        help='Set defaults for --tag-prefix and --git-dir for the notmuch bug tracker')
    parser.add_argument(
        '-l', '--log-level',
        choices=['critical', 'error', 'warning', 'info', 'debug'],
        help='Log verbosity.  Defaults to {!r}.'.format(
            _logging.getLevelName(_LOG.level).lower()))

    help = _functools.partial(_help, parser=parser)
    help.__doc__ = _help.__doc__
    subparsers = parser.add_subparsers(
        title='commands',
        description=(
            'For help on a particular command, run: '
            "'%(prog)s ... <command> --help'."))
    for command in [
            'archive',
            'checkout',
            'clone',
            'commit',
            'fetch',
            'help',
            'init',
            'log',
            'merge',
            'pull',
            'push',
            'status',
            ]:
        func = locals()[command]
        doc = _textwrap.dedent(func.__doc__).strip().replace('%', '%%')
        subparser = subparsers.add_parser(
            command,
            help=doc.splitlines()[0],
            description=doc,
            formatter_class=argparse.RawDescriptionHelpFormatter)
        subparser.set_defaults(func=func)
        if command == 'archive':
            subparser.add_argument(
                'treeish', metavar='TREE-ISH', nargs='?', default='HEAD',
                help=(
                    'The tree or commit to produce an archive for.  Defaults '
                    "to 'HEAD'."))
            subparser.add_argument(
                'args', metavar='ARG', nargs='*',
                help=(
                    "Argument passed through to 'git archive'.  Set anything "
                    'before <tree-ish>, see git-archive(1) for details.'))
        elif command == 'checkout':
            subparser.add_argument(
                '-f', '--force', action='store_true',
                help='checkout a large fraction of tags.')
        elif command == 'clone':
            subparser.add_argument(
                'repository',
                help=(
                    'The (possibly remote) repository to clone from.  See the '
                    'URLS section of git-clone(1) for more information on '
                    'specifying repositories.'))
        elif command == 'commit':
            subparser.add_argument(
                '-f', '--force', action='store_true',
                help='commit a large fraction of tags.')
            subparser.add_argument(
                'message', metavar='MESSAGE', default='', nargs='?',
                help='Text for the commit message.')
        elif command == 'fetch':
            subparser.add_argument(
                'remote', metavar='REMOTE', nargs='?',
                help=(
                    'Override the default configured in branch.<name>.remote '
                    'to fetch from a particular remote repository (e.g. '
                    "'origin')."))
        elif command == 'help':
            subparser.add_argument(
                'command', metavar='COMMAND', nargs='?',
                help='The command to show help for.')
        elif command == 'init':
            subparser.add_argument(
                '--format-version', metavar='VERSION',
                default = None,
                help='create format VERSION repository.')
        elif command == 'log':
            subparser.add_argument(
                'args', metavar='ARG', nargs='*',
                help="Additional argument passed through to 'git log'.")
        elif command == 'merge':
            subparser.add_argument(
                'reference', metavar='REFERENCE', default='@{upstream}',
                nargs='?',
                help=(
                    'Reference, usually other branch heads, to merge into '
                    "our branch.  Defaults to '@{upstream}'."))
        elif command == 'pull':
            subparser.add_argument(
                'repository', metavar='REPOSITORY', default=None, nargs='?',
                help=(
                    'The "remote" repository that is the source of the pull.  '
                    'This parameter can be either a URL (see the section GIT '
                    'URLS in git-pull(1)) or the name of a remote (see the '
                    'section REMOTES in git-pull(1)).'))
            subparser.add_argument(
                'refspecs', metavar='REFSPEC', default=None, nargs='*',
                help=(
                    'Refspec (usually a branch name) to fetch and merge.  See '
                    'the <refspec> entry in the OPTIONS section of '
                    'git-pull(1) for other possibilities.'))
        elif command == 'push':
            subparser.add_argument(
               'repository', metavar='REPOSITORY', default=None, nargs='?',
                help=(
                    'The "remote" repository that is the destination of the '
                    'push.  This parameter can be either a URL (see the '
                    'section GIT URLS in git-push(1)) or the name of a remote '
                    '(see the section REMOTES in git-push(1)).'))
            subparser.add_argument(
                'refspecs', metavar='REFSPEC', default=None, nargs='*',
                help=(
                    'Refspec (usually a branch name) to push.  See '
                    'the <refspec> entry in the OPTIONS section of '
                    'git-push(1) for other possibilities.'))

    args = parser.parse_args()

    nmbug_mode = False
    notmuch_profile = _os.getenv('NOTMUCH_PROFILE','default')

    if args.nmbug or _os.path.basename(__file__) == 'nmbug':
        nmbug_mode = True

    if args.git_dir:
        NOTMUCH_GIT_DIR = args.git_dir
    else:
        if nmbug_mode:
            default = _os.path.join('~', '.nmbug')
        else:
            default = _notmuch_config_get ('git.path')
            if default == '':
                default = xdg_data_path(notmuch_profile)

        NOTMUCH_GIT_DIR = _os.path.expanduser(_os.getenv('NOTMUCH_GIT_DIR', default))

    _NOTMUCH_GIT_DIR = _os.path.join(NOTMUCH_GIT_DIR, '.git')
    if _os.path.isdir(_NOTMUCH_GIT_DIR):
        NOTMUCH_GIT_DIR = _NOTMUCH_GIT_DIR

    if args.tag_prefix:
        TAG_PREFIX = args.tag_prefix
    else:
        if nmbug_mode:
            prefix = 'notmuch::'
        else:
            prefix = _notmuch_config_get ('git.tag_prefix')

        TAG_PREFIX =  _os.getenv('NOTMUCH_GIT_PREFIX', prefix)

    _ENCODED_TAG_PREFIX = _hex_quote(TAG_PREFIX, safe='+@=,')  # quote ':'

    if args.log_level:
        level = getattr(_logging, args.log_level.upper())
        _LOG.setLevel(level)

    # for test suite
    for var in ['NOTMUCH_GIT_DIR', 'NOTMUCH_GIT_PREFIX', 'NOTMUCH_PROFILE', 'NOTMUCH_CONFIG' ]:
        _LOG.debug('env {:s} = {:s}'.format(var, _os.getenv(var,'%None%')))

    if _notmuch_config_get('built_with.sexp_queries') != 'true':
        _LOG.error("notmuch git needs sexp query support")
        _sys.exit(1)

    if not getattr(args, 'func', None):
        parser.print_usage()
        _sys.exit(1)

    # The following two lines are used by the test suite.
    _LOG.debug('prefix = {:s}'.format(TAG_PREFIX))
    _LOG.debug('repository = {:s}'.format(NOTMUCH_GIT_DIR))

    if args.func != init:
        FORMAT_VERSION = read_format_version()

    _LOG.debug('FORMAT_VERSION={:d}'.format(FORMAT_VERSION))

    if args.func == help:
        arg_names = ['command']
    else:
        (arg_names, varargs, varkw) = _inspect.getargs(args.func.__code__)
    kwargs = {key: getattr(args, key) for key in arg_names if key in args}
    try:
        args.func(**kwargs)
    except SubprocessError as e:
        if _LOG.level == _logging.DEBUG:
            raise  # don't mask the traceback
        _LOG.error(str(e))
        _sys.exit(1)



#!/usr/bin/env python3

# to launch nmweb from gunicorn.

from nmweb import urls, index, search, show
import web

app = web.application(urls, globals())

# get the wsgi app from web.py application object
wsgiapp = app.wsgifunc()



import collections
import contextlib
import os
import pathlib
import weakref

import notmuch2._base as base
import notmuch2._capi as capi
import notmuch2._errors as errors
import notmuch2._tags as tags


__all__ = ['Message']


class Message(base.NotmuchObject):
    """An email message stored in the notmuch database retrieved via a query.

    This should not be directly created, instead it will be returned
    by calling methods on :class:`Database`.  A message keeps a
    reference to the database object since the database object can not
    be released while the message is in use.

    Note that this represents a message in the notmuch database.  For
    full email functionality you may want to use the :mod:`email`
    package from Python's standard library.  You could e.g. create
    this as such::

       notmuch_msg = db.find(msgid)  # or from a query
       parser = email.parser.BytesParser(policy=email.policy.default)
       with notmuch_msg.path.open('rb) as fp:
           email_msg = parser.parse(fp)

    Most commonly the functionality provided by notmuch is sufficient
    to read email however.

    Messages are considered equal when they have the same message ID.
    This is how libnotmuch treats messages as well, the
    :meth:`pathnames` function returns multiple results for
    duplicates.

    :param parent: The parent object.  This is probably one off a
       :class:`Database`, :class:`Thread` or :class:`Query`.
    :type parent: NotmuchObject
    :param db: The database instance this message is associated with.
       This could be the same as the parent.
    :type db: Database
    :param msg_p: The C pointer to the ``notmuch_message_t``.
    :type msg_p: <cdata>
    :param dup: Whether the message was a duplicate on insertion.
    :type dup: None or bool
    """
    _msg_p = base.MemoryPointer()

    def __init__(self, parent, msg_p, *, db):
        self._parent = parent
        self._msg_p = msg_p
        self._db = db

    @property
    def alive(self):
        if not self._parent.alive:
            return False
        try:
            self._msg_p
        except errors.ObjectDestroyedError:
            return False
        else:
            return True

    def __del__(self):
        self._destroy()

    def _destroy(self):
        if self.alive:
            capi.lib.notmuch_message_destroy(self._msg_p)
        self._msg_p = None

    @property
    def messageid(self):
        """The message ID as a string.

        The message ID is decoded with the ignore error handler.  This
        is fine as long as the message ID is well formed.  If it is
        not valid ASCII then this will be lossy.  So if you need to be
        able to write the exact same message ID back you should use
        :attr:`messageidb`.

        Note that notmuch will decode the message ID value and thus
        strip off the surrounding ``<`` and ``>`` characters.  This is
        different from Python's :mod:`email` package behaviour which
        leaves these characters in place.

        :returns: The message ID.
        :rtype: :class:`BinString`, this is a normal str but calling
           bytes() on it will return the original bytes used to create
           it.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_message_id(self._msg_p)
        return base.BinString(capi.ffi.string(ret))

    @property
    def threadid(self):
        """The thread ID.

        The thread ID is decoded with the surrogateescape error
        handler so that it is possible to reconstruct the original
        thread ID if it is not valid UTF-8.

        :returns: The thread ID.
        :rtype: :class:`BinString`, this is a normal str but calling
           bytes() on it will return the original bytes used to create
           it.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_thread_id(self._msg_p)
        return base.BinString(capi.ffi.string(ret))

    @property
    def path(self):
        """A pathname of the message as a pathlib.Path instance.

        If multiple files in the database contain the same message ID
        this will be just one of the files, chosen at random.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_filename(self._msg_p)
        return pathlib.Path(os.fsdecode(capi.ffi.string(ret)))

    @property
    def pathb(self):
        """A pathname of the message as a bytes object.

        See :attr:`path` for details, this is the same but does return
        the path as a bytes object which is faster but less convenient.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_filename(self._msg_p)
        return capi.ffi.string(ret)

    def filenames(self):
        """Return an iterator of all files for this message.

        If multiple files contained the same message ID they will all
        be returned here.  The files are returned as instances of
        :class:`pathlib.Path`.

        :returns: Iterator yielding :class:`pathlib.Path` instances.
        :rtype: iter

        :raises ObjectDestroyedError: if used after destroyed.
        """
        fnames_p = capi.lib.notmuch_message_get_filenames(self._msg_p)
        return PathIter(self, fnames_p)

    def filenamesb(self):
        """Return an iterator of all files for this message.

        This is like :meth:`pathnames` but the files are returned as
        byte objects instead.

        :returns: Iterator yielding :class:`bytes` instances.
        :rtype: iter

        :raises ObjectDestroyedError: if used after destroyed.
        """
        fnames_p = capi.lib.notmuch_message_get_filenames(self._msg_p)
        return FilenamesIter(self, fnames_p)

    @property
    def ghost(self):
        """Indicates whether this message is a ghost message.

        A ghost message if a message which we know exists, but it has
        no files or content associated with it.  This can happen if
        it was referenced by some other message.  Only the
        :attr:`messageid` and :attr:`threadid` attributes are valid
        for it.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_flag(
            self._msg_p, capi.lib.NOTMUCH_MESSAGE_FLAG_GHOST)
        return bool(ret)

    @property
    def excluded(self):
        """Indicates whether this message was excluded from the query.

        When a message is created from a search, sometimes messages
        that where excluded by the search query could still be
        returned by it, e.g. because they are part of a thread
        matching the query.  the :meth:`Database.query` method allows
        these messages to be flagged, which results in this property
        being set to *True*.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_flag(
            self._msg_p, capi.lib.NOTMUCH_MESSAGE_FLAG_EXCLUDED)
        return bool(ret)

    @property
    def matched(self):
        """Indicates whether this message was matched by the query.

        When a thread is created from a search, some of the
        messages may not match the original query.  This property
        is set to *True* for those that do match.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_get_flag(
            self._msg_p, capi.lib.NOTMUCH_MESSAGE_FLAG_MATCH)
        return bool(ret)

    @property
    def date(self):
        """The message date as an integer.

        The time the message was sent as an integer number of seconds
        since the *epoch*, 1 Jan 1970.  This is derived from the
        message's header, you can get the original header value with
        :meth:`header`.

        :raises ObjectDestroyedError: if used after destroyed.
        """
        return capi.lib.notmuch_message_get_date(self._msg_p)

    def header(self, name):
        """Return the value of the named header.

        Returns the header from notmuch, some common headers are
        stored in the database, others are read from the file.
        Headers are returned with their newlines stripped and
        collapsed concatenated together if they occur multiple times.
        You may be better off using the standard library email
        package's ``email.message_from_file(msg.path.open())`` if that
        is not sufficient for you.

        :param header: Case-insensitive header name to retrieve.
        :type header: str or bytes

        :returns: The header value, an empty string if the header is
           not present.
        :rtype: str

        :raises LookupError: if the header is not present.
        :raises NullPointerError: For unexpected notmuch errors.
        :raises ObjectDestroyedError: if used after destroyed.
        """
        # The returned is supposedly guaranteed to be UTF-8.  Header
        # names must be ASCII as per RFC x822.
        if isinstance(name, str):
            name = name.encode('ascii')
        ret = capi.lib.notmuch_message_get_header(self._msg_p, name)
        if ret == capi.ffi.NULL:
            raise errors.NullPointerError()
        hdr = capi.ffi.string(ret)
        if not hdr:
            raise LookupError
        return hdr.decode(encoding='utf-8')

    @property
    def tags(self):
        """The tags associated with the message.

        This behaves as a set.  But removing and adding items to the
        set removes and adds them to the message in the database.

        :raises ReadOnlyDatabaseError: When manipulating tags on a
           database opened in read-only mode.
        :raises ObjectDestroyedError: if used after destroyed.
        """
        try:
            ref = self._cached_tagset
        except AttributeError:
            tagset = None
        else:
            tagset = ref()
        if tagset is None:
            tagset = tags.MutableTagSet(
                self, '_msg_p', capi.lib.notmuch_message_get_tags)
            self._cached_tagset = weakref.ref(tagset)
        return tagset

    @contextlib.contextmanager
    def frozen(self):
        """Context manager to freeze the message state.

        This allows you to perform atomic tag updates::

           with msg.frozen():
               msg.tags.clear()
               msg.tags.add('foo')

        Using This would ensure the message never ends up with no tags
        applied at all.

        It is safe to nest calls to this context manager.

        :raises ReadOnlyDatabaseError: if the database is opened in
           read-only mode.
        :raises UnbalancedFreezeThawError: if you somehow managed to
           call __exit__ of this context manager more than once.  Why
           did you do that?
        :raises ObjectDestroyedError: if used after destroyed.
        """
        ret = capi.lib.notmuch_message_freeze(self._msg_p)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)
        self._frozen = True
        try:
            yield
        except Exception:
            # Only way to "rollback" these changes is to destroy
            # ourselves and re-create.  Behold.
            msgid = self.messageid
            self._destroy()
            with contextlib.suppress(Exception):
                new = self._db.find(msgid)
                self._msg_p = new._msg_p
                new._msg_p = None
                del new
            raise
        else:
            ret = capi.lib.notmuch_message_thaw(self._msg_p)
            if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
                raise errors.NotmuchError(ret)
            self._frozen = False

    @property
    def properties(self):
        """A map of arbitrary key-value pairs associated with the message.

        Be aware that properties may be used by other extensions to
        store state in.  So delete or modify with care.

        The properties map is somewhat special.  It is essentially a
        multimap-like structure where each key can have multiple
        values.  Therefore accessing a single item using
        :meth:`PropertiesMap.get` or :meth:`PropertiesMap.__getitem__`
        will only return you the *first* item if there are multiple
        and thus are only recommended if you know there to be only one
        value.

        Instead the map has an additional :meth:`PropertiesMap.all`
        method which can be used to retrieve all properties of a given
        key.  This method also allows iterating of a a subset of the
        keys starting with a given prefix.
        """
        try:
            ref = self._cached_props
        except AttributeError:
            props = None
        else:
            props = ref()
        if props is None:
            props = PropertiesMap(self, '_msg_p')
            self._cached_props = weakref.ref(props)
        return props

    def replies(self):
        """Return an iterator of all replies to this message.

        This method will only work if the message was created from a
        thread.  Otherwise it will yield no results.

        :returns: An iterator yielding :class:`OwnedMessage` instances.
        :rtype: MessageIter
        """
        # The notmuch_messages_valid call accepts NULL and this will
        # become an empty iterator, raising StopIteration immediately.
        # Hence no return value checking here.
        msgs_p = capi.lib.notmuch_message_get_replies(self._msg_p)
        return MessageIter(self, msgs_p, db=self._db, msg_cls=OwnedMessage)

    def __hash__(self):
        return hash(self.messageid)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.messageid == other.messageid


class OwnedMessage(Message):
    """An email message owned by parent thread object.

    This subclass of Message is used for messages that are retrieved
    from the notmuch database via a parent :class:`notmuch2.Thread`
    object, which "owns" this message.  This means that when this
    message object is destroyed, by calling :func:`del` or
    :meth:`_destroy` directly or indirectly, the message is not freed
    in the notmuch API and the parent :class:`notmuch2.Thread` object
    can return the same object again when needed.
    """

    @property
    def alive(self):
        return self._parent.alive

    def _destroy(self):
        pass


class FilenamesIter(base.NotmuchIter):
    """Iterator for binary filenames objects."""

    def __init__(self, parent, iter_p):
        super().__init__(parent, iter_p,
                         fn_destroy=capi.lib.notmuch_filenames_destroy,
                         fn_valid=capi.lib.notmuch_filenames_valid,
                         fn_get=capi.lib.notmuch_filenames_get,
                         fn_next=capi.lib.notmuch_filenames_move_to_next)

    def __next__(self):
        fname = super().__next__()
        return capi.ffi.string(fname)


class PathIter(FilenamesIter):
    """Iterator for pathlib.Path objects."""

    def __next__(self):
        fname = super().__next__()
        return pathlib.Path(os.fsdecode(fname))


class PropertiesMap(base.NotmuchObject, collections.abc.MutableMapping):
    """A mutable mapping to manage properties.

    Both keys and values of properties are supposed to be UTF-8
    strings in libnotmuch.  However since the uderlying API uses
    bytestrings you can use either str or bytes to represent keys and
    all returned keys and values use :class:`BinString`.

    Also be aware that ``iter(this_map)`` will return duplicate keys,
    while the :class:`collections.abc.KeysView` returned by
    :meth:`keys` is a :class:`collections.abc.Set` subclass.  This
    means the former will yield duplicate keys while the latter won't.
    It also means ``len(list(iter(this_map)))`` could be different
    than ``len(this_map.keys())``.  ``len(this_map)`` will correspond
    with the length of the default iterator.

    Be aware that libnotmuch exposes all of this as iterators, so
    quite a few operations have O(n) performance instead of the usual
    O(1).
    """
    Property = collections.namedtuple('Property', ['key', 'value'])
    _marker = object()

    def __init__(self, msg, ptr_name):
        self._msg = msg
        self._ptr = lambda: getattr(msg, ptr_name)

    @property
    def alive(self):
        if not self._msg.alive:
            return False
        try:
            self._ptr
        except errors.ObjectDestroyedError:
            return False
        else:
            return True

    def _destroy(self):
        pass

    def __iter__(self):
        """Return an iterator which iterates over the keys.

        Be aware that a single key may have multiple values associated
        with it, if so it will appear multiple times here.
        """
        iter_p = capi.lib.notmuch_message_get_properties(self._ptr(), b'', 0)
        return PropertiesKeyIter(self, iter_p)

    def __len__(self):
        iter_p = capi.lib.notmuch_message_get_properties(self._ptr(), b'', 0)
        it = base.NotmuchIter(
            self, iter_p,
            fn_destroy=capi.lib.notmuch_message_properties_destroy,
            fn_valid=capi.lib.notmuch_message_properties_valid,
            fn_get=capi.lib.notmuch_message_properties_key,
            fn_next=capi.lib.notmuch_message_properties_move_to_next,
        )
        return len(list(it))

    def __getitem__(self, key):
        """Return **the first** peroperty associated with a key."""
        if isinstance(key, str):
            key = key.encode('utf-8')
        value_pp = capi.ffi.new('char**')
        ret = capi.lib.notmuch_message_get_property(self._ptr(), key, value_pp)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)
        if value_pp[0] == capi.ffi.NULL:
            raise KeyError
        return base.BinString.from_cffi(value_pp[0])

    def keys(self):
        """Return a :class:`collections.abc.KeysView` for this map.

        Even when keys occur multiple times this is a subset of set()
        so will only contain them once.
        """
        return collections.abc.KeysView({k: None for k in self})

    def items(self):
        """Return a :class:`collections.abc.ItemsView` for this map.

        The ItemsView treats a ``(key, value)`` pair as unique, so
        dupcliate ``(key, value)`` pairs will be merged together.
        However duplicate keys with different values will be returned.
        """
        items = set()
        props_p = capi.lib.notmuch_message_get_properties(self._ptr(), b'', 0)
        while capi.lib.notmuch_message_properties_valid(props_p):
            key = capi.lib.notmuch_message_properties_key(props_p)
            value = capi.lib.notmuch_message_properties_value(props_p)
            items.add((base.BinString.from_cffi(key),
                       base.BinString.from_cffi(value)))
            capi.lib.notmuch_message_properties_move_to_next(props_p)
        capi.lib.notmuch_message_properties_destroy(props_p)
        return PropertiesItemsView(items)

    def values(self):
        """Return a :class:`collecions.abc.ValuesView` for this map.

        All unique property values are included in the view.
        """
        values = set()
        props_p = capi.lib.notmuch_message_get_properties(self._ptr(), b'', 0)
        while capi.lib.notmuch_message_properties_valid(props_p):
            value = capi.lib.notmuch_message_properties_value(props_p)
            values.add(base.BinString.from_cffi(value))
            capi.lib.notmuch_message_properties_move_to_next(props_p)
        capi.lib.notmuch_message_properties_destroy(props_p)
        return PropertiesValuesView(values)

    def __setitem__(self, key, value):
        """Add a key-value pair to the properties.

        You may prefer to use :meth:`add` for clarity since this
        method usually implies implicit overwriting of an existing key
        if it exists, while for properties this is not the case.
        """
        self.add(key, value)

    def add(self, key, value):
        """Add a key-value pair to the properties."""
        if isinstance(key, str):
            key = key.encode('utf-8')
        if isinstance(value, str):
            value = value.encode('utf-8')
        ret = capi.lib.notmuch_message_add_property(self._ptr(), key, value)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def __delitem__(self, key):
        """Remove all properties with this key."""
        if isinstance(key, str):
            key = key.encode('utf-8')
        ret = capi.lib.notmuch_message_remove_all_properties(self._ptr(), key)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def remove(self, key, value):
        """Remove a key-value pair from the properties."""
        if isinstance(key, str):
            key = key.encode('utf-8')
        if isinstance(value, str):
            value = value.encode('utf-8')
        ret = capi.lib.notmuch_message_remove_property(self._ptr(), key, value)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def pop(self, key, default=_marker):
        try:
            value = self[key]
        except KeyError:
            if default is self._marker:
                raise
            else:
                return default
        else:
            self.remove(key, value)
            return value

    def popitem(self):
        try:
            key = next(iter(self))
        except StopIteration:
            raise KeyError
        value = self.pop(key)
        return (key, value)

    def clear(self):
        ret = capi.lib.notmuch_message_remove_all_properties(self._ptr(),
                                                             capi.ffi.NULL)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def getall(self, prefix='', *, exact=False):
        """Return an iterator yielding all properties for a given key prefix.

        The returned iterator yields all peroperties which start with
        a given key prefix as ``(key, value)`` namedtuples.  If called
        with ``exact=True`` then only properties which exactly match
        the prefix are returned, those a key longer than the prefix
        will not be included.

        :param prefix: The prefix of the key.
        """
        if isinstance(prefix, str):
            prefix = prefix.encode('utf-8')
        props_p = capi.lib.notmuch_message_get_properties(self._ptr(),
                                                          prefix, exact)
        return PropertiesIter(self, props_p)


class PropertiesKeyIter(base.NotmuchIter):

    def __init__(self, parent, iter_p):
        super().__init__(
            parent,
            iter_p,
            fn_destroy=capi.lib.notmuch_message_properties_destroy,
            fn_valid=capi.lib.notmuch_message_properties_valid,
            fn_get=capi.lib.notmuch_message_properties_key,
            fn_next=capi.lib.notmuch_message_properties_move_to_next)

    def __next__(self):
        item = super().__next__()
        return base.BinString.from_cffi(item)


class PropertiesIter(base.NotmuchIter):

    def __init__(self, parent, iter_p):
        super().__init__(
            parent,
            iter_p,
            fn_destroy=capi.lib.notmuch_message_properties_destroy,
            fn_valid=capi.lib.notmuch_message_properties_valid,
            fn_get=capi.lib.notmuch_message_properties_key,
            fn_next=capi.lib.notmuch_message_properties_move_to_next,
        )

    def __next__(self):
        if not self._fn_valid(self._iter_p):
            self._destroy()
            raise StopIteration
        key = capi.lib.notmuch_message_properties_key(self._iter_p)
        value = capi.lib.notmuch_message_properties_value(self._iter_p)
        capi.lib.notmuch_message_properties_move_to_next(self._iter_p)
        return PropertiesMap.Property(base.BinString.from_cffi(key),
                                      base.BinString.from_cffi(value))


class PropertiesItemsView(collections.abc.Set):

    __slots__ = ('_items',)

    def __init__(self, items):
        self._items = items

    @classmethod
    def _from_iterable(self, it):
        return set(it)

    def __len__(self):
        return len(self._items)

    def __contains__(self, item):
        return item in self._items

    def __iter__(self):
        yield from self._items


collections.abc.ItemsView.register(PropertiesItemsView)


class PropertiesValuesView(collections.abc.Set):

    __slots__ = ('_values',)

    def __init__(self, values):
        self._values = values

    def __len__(self):
        return len(self._values)

    def __contains__(self, value):
        return value in self._values

    def __iter__(self):
        yield from self._values


collections.abc.ValuesView.register(PropertiesValuesView)


class MessageIter(base.NotmuchIter):

    def __init__(self, parent, msgs_p, *, db, msg_cls=Message):
        self._db = db
        self._msg_cls = msg_cls
        super().__init__(parent, msgs_p,
                         fn_destroy=capi.lib.notmuch_messages_destroy,
                         fn_valid=capi.lib.notmuch_messages_valid,
                         fn_get=capi.lib.notmuch_messages_get,
                         fn_next=capi.lib.notmuch_messages_move_to_next)

    def __next__(self):
        msg_p = super().__next__()
        return self._msg_cls(self, msg_p, db=self._db)



import collections
import configparser
import os
import pathlib

import pytest

import notmuch2
import notmuch2._errors as errors
import notmuch2._database as dbmod
import notmuch2._message as message


@pytest.fixture
def db(maildir):
    with dbmod.Database.create(maildir.path, config=notmuch2.Database.CONFIG.EMPTY) as db:
        yield db


class TestDefaultDb:
    """Tests for reading the default database.

    The error cases are fairly undefined, some relevant Python error
    will come out if you give it a bad filename or if the file does
    not parse correctly.  So we're not testing this too deeply.
    """

    def test_config_pathname_default(self, monkeypatch):
        monkeypatch.delenv('NOTMUCH_CONFIG', raising=False)
        user = pathlib.Path('~/.notmuch-config').expanduser()
        assert dbmod._config_pathname() == user

    def test_config_pathname_env(self, monkeypatch):
        monkeypatch.setenv('NOTMUCH_CONFIG', '/some/random/path')
        assert dbmod._config_pathname() == pathlib.Path('/some/random/path')

    def test_default_path_nocfg(self, monkeypatch, tmppath):
        monkeypatch.setenv('NOTMUCH_CONFIG', str(tmppath/'foo'))
        with pytest.raises(FileNotFoundError):
            dbmod.Database.default_path()

    def test_default_path_cfg_is_dir(self, monkeypatch, tmppath):
        monkeypatch.setenv('NOTMUCH_CONFIG', str(tmppath))
        with pytest.raises(IsADirectoryError):
            dbmod.Database.default_path()

    def test_default_path_parseerr(self, monkeypatch, tmppath):
        cfg = tmppath / 'notmuch-config'
        with cfg.open('w') as fp:
            fp.write('invalid')
        monkeypatch.setenv('NOTMUCH_CONFIG', str(cfg))
        with pytest.raises(configparser.Error):
            dbmod.Database.default_path()

    def test_default_path_parse(self, monkeypatch, tmppath):
        cfg = tmppath / 'notmuch-config'
        with cfg.open('w') as fp:
            fp.write('[database]\n')
            fp.write('path={!s}'.format(tmppath))
        monkeypatch.setenv('NOTMUCH_CONFIG', str(cfg))
        assert dbmod.Database.default_path() == tmppath

    def test_default_path_param(self, monkeypatch, tmppath):
        cfg_dummy = tmppath / 'dummy'
        monkeypatch.setenv('NOTMUCH_CONFIG', str(cfg_dummy))
        cfg_real = tmppath / 'notmuch_config'
        with cfg_real.open('w') as fp:
            fp.write('[database]\n')
            fp.write('path={!s}'.format(cfg_real/'mail'))
        assert dbmod.Database.default_path(cfg_real) == cfg_real/'mail'


class TestCreate:

    def test_create(self, tmppath, db):
        assert tmppath.joinpath('.notmuch/xapian/').exists()

    def test_create_already_open(self, tmppath, db):
        with pytest.raises(errors.NotmuchError):
            db.create(tmppath)

    def test_create_existing(self, tmppath, db):
        with pytest.raises(errors.DatabaseExistsError):
            dbmod.Database.create(path=tmppath)

    def test_close(self, db):
        db.close()

    def test_del_noclose(self, db):
        del db

    def test_close_del(self, db):
        db.close()
        del db

    def test_closed_attr(self, db):
        assert not db.closed
        db.close()
        assert db.closed

    def test_ctx(self, db):
        with db as ctx:
            assert ctx is db
            assert not db.closed
        assert db.closed

    def test_path(self, db, tmppath):
        assert db.path == tmppath

    def test_version(self, db):
        assert db.version > 0

    def test_needs_upgrade(self, db):
        assert db.needs_upgrade in (True, False)


class TestAtomic:

    def test_exit_early(self, db):
        with pytest.raises(errors.UnbalancedAtomicError):
            with db.atomic() as ctx:
                ctx.force_end()

    def test_exit_late(self, db):
        with db.atomic() as ctx:
            pass
        with pytest.raises(errors.UnbalancedAtomicError):
            ctx.force_end()

    def test_abort(self, db):
        with db.atomic() as txn:
            txn.abort()
        assert db.closed


class TestRevision:

    def test_single_rev(self, db):
        r = db.revision()
        assert isinstance(r, dbmod.DbRevision)
        assert isinstance(r.rev, int)
        assert isinstance(r.uuid, bytes)
        assert r is r
        assert r == r
        assert r <= r
        assert r >= r
        assert not r < r
        assert not r > r

    def test_diff_db(self, tmppath):
        dbpath0 = tmppath.joinpath('db0')
        dbpath0.mkdir()
        dbpath1 = tmppath.joinpath('db1')
        dbpath1.mkdir()
        db0 = dbmod.Database.create(path=dbpath0)
        db1 = dbmod.Database.create(path=dbpath1)
        r_db0 = db0.revision()
        r_db1 = db1.revision()
        assert r_db0 != r_db1
        assert r_db0.uuid != r_db1.uuid

    def test_cmp(self, db, maildir):
        rev0 = db.revision()
        _, pathname = maildir.deliver()
        db.add(pathname, sync_flags=False)
        rev1 = db.revision()
        assert rev0 < rev1
        assert rev0 <= rev1
        assert not rev0 > rev1
        assert not rev0 >= rev1
        assert not rev0 == rev1
        assert rev0 != rev1

    # XXX add tests for revisions comparisons

class TestMessages:

    def test_add_message(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(pathname, sync_flags=False)
        assert isinstance(msg, message.Message)
        assert msg.path == pathname
        assert msg.messageid == msgid

    def test_add_message_str(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(str(pathname), sync_flags=False)

    def test_add_message_bytes(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(os.fsencode(bytes(pathname)), sync_flags=False)

    def test_remove_message(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(pathname, sync_flags=False)
        assert db.find(msgid)
        dup = db.remove(pathname)
        with pytest.raises(LookupError):
            db.find(msgid)

    def test_remove_message_str(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(pathname, sync_flags=False)
        assert db.find(msgid)
        dup = db.remove(str(pathname))
        with pytest.raises(LookupError):
            db.find(msgid)

    def test_remove_message_bytes(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg, dup = db.add(pathname, sync_flags=False)
        assert db.find(msgid)
        dup = db.remove(os.fsencode(bytes(pathname)))
        with pytest.raises(LookupError):
            db.find(msgid)

    def test_find_message(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg0, dup = db.add(pathname, sync_flags=False)
        msg1 = db.find(msgid)
        assert isinstance(msg1, message.Message)
        assert msg1.messageid == msgid == msg0.messageid
        assert msg1.path == pathname == msg0.path

    def test_find_message_notfound(self, db):
        with pytest.raises(LookupError):
            db.find('foo')

    def test_get_message(self, db, maildir):
        msgid, pathname = maildir.deliver()
        msg0, _ = db.add(pathname, sync_flags=False)
        msg1 = db.get(pathname)
        assert isinstance(msg1, message.Message)
        assert msg1.messageid == msgid == msg0.messageid
        assert msg1.path == pathname == msg0.path

    def test_get_message_str(self, db, maildir):
        msgid, pathname = maildir.deliver()
        db.add(pathname, sync_flags=False)
        msg = db.get(str(pathname))
        assert msg.messageid == msgid

    def test_get_message_bytes(self, db, maildir):
        msgid, pathname = maildir.deliver()
        db.add(pathname, sync_flags=False)
        msg = db.get(os.fsencode(bytes(pathname)))
        assert msg.messageid == msgid


class TestTags:
    # We just want to test this behaves like a set at a hight level.
    # The set semantics are tested in detail in the test_tags module.

    def test_type(self, db):
        assert isinstance(db.tags, collections.abc.Set)

    def test_none(self, db):
        itags = iter(db.tags)
        with pytest.raises(StopIteration):
            next(itags)
        assert len(db.tags) == 0
        assert not db.tags

    def test_some(self, db, maildir):
        _, pathname = maildir.deliver()
        msg, _ = db.add(pathname, sync_flags=False)
        msg.tags.add('hello')
        itags = iter(db.tags)
        assert next(itags) == 'hello'
        with pytest.raises(StopIteration):
            next(itags)
        assert 'hello' in msg.tags

    def test_cache(self, db):
        assert db.tags is db.tags

    def test_iters(self, db):
        i1 = iter(db.tags)
        i2 = iter(db.tags)
        assert i1 is not i2


class TestQuery:

    @pytest.fixture
    def db(self, maildir, notmuch):
        """Return a read-only notmuch2.Database.

        The database will have 3 messages, 2 threads.
        """
        msgid, _ = maildir.deliver(body='foo')
        maildir.deliver(body='bar')
        maildir.deliver(body='baz',
                        headers=[('In-Reply-To', '<{}>'.format(msgid))])
        notmuch('new')
        with dbmod.Database(maildir.path, 'rw', config=notmuch2.Database.CONFIG.EMPTY) as db:
            yield db

    def test_count_messages(self, db):
        assert db.count_messages('*') == 3

    def test_messages_type(self, db):
        msgs = db.messages('*')
        assert isinstance(msgs, collections.abc.Iterator)

    def test_message_no_results(self, db):
        msgs = db.messages('not_a_matching_query')
        with pytest.raises(StopIteration):
            next(msgs)

    def test_message_match(self, db):
        msgs = db.messages('*')
        msg = next(msgs)
        assert isinstance(msg, notmuch2.Message)

    def test_count_threads(self, db):
        assert db.count_threads('*') == 2

    def test_threads_type(self, db):
        threads = db.threads('*')
        assert isinstance(threads, collections.abc.Iterator)

    def test_threads_no_match(self, db):
        threads = db.threads('not_a_matching_query')
        with pytest.raises(StopIteration):
            next(threads)

    def test_threads_match(self, db):
        threads = db.threads('*')
        thread = next(threads)
        assert isinstance(thread, notmuch2.Thread)

    def test_use_threaded_message_twice(self, db):
        thread = next(db.threads('*'))
        for msg in thread.toplevel():
            assert isinstance(msg, notmuch2.Message)
            assert msg.alive
            del msg
        for msg in thread:
            assert isinstance(msg, notmuch2.Message)
            assert msg.alive
            del msg



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from .globals import (
    nmlib,
    Python3StringMixIn,
    NotmuchThreadP,
    NotmuchThreadsP,
)
from .errors import (
    NullPointerError,
    NotInitializedError,
)
from .thread import Thread

class Threads(Python3StringMixIn):
    """Represents a list of notmuch threads

    This object provides an iterator over a list of notmuch threads
    (Technically, it provides a wrapper for the underlying
    *notmuch_threads_t* structure). Do note that the underlying
    library only provides a one-time iterator (it cannot reset the
    iterator to the start). Thus iterating over the function will
    "exhaust" the list of threads, and a subsequent iteration attempt
    will raise a :exc:`NotInitializedError`. Also
    note, that any function that uses iteration will also
    exhaust the messages. So both::

      for thread in threads: print thread

    as well as::

       list_of_threads = list(threads)

    will "exhaust" the threads. If you need to re-iterate over a list of
    messages you will need to retrieve a new :class:`Threads` object.

    Things are not as bad as it seems though, you can store and reuse
    the single Thread objects as often as you want as long as you
    keep the parent Threads object around. (Recall that due to
    hierarchical memory allocation, all derived Threads objects will
    be invalid when we delete the parent Threads() object, even if it
    was already "exhausted".) So this works::

      db   = Database()
      threads = Query(db,'').search_threads() #get a Threads() object
      threadlist = []
      for thread in threads:
         threadlist.append(thread)

      # threads is "exhausted" now.
      # However it will be kept around until all retrieved Thread() objects are
      # also deleted. If you did e.g. an explicit del(threads) here, the
      # following lines would fail.

      # You can reiterate over *threadlist* however as often as you want.
      # It is simply a list with Thread objects.

      print (threadlist[0].get_thread_id())
      print (threadlist[1].get_thread_id())
      print (threadlist[0].get_total_messages())
    """

    #notmuch_threads_get
    _get = nmlib.notmuch_threads_get
    _get.argtypes = [NotmuchThreadsP]
    _get.restype = NotmuchThreadP

    def __init__(self, threads_p, parent=None):
        """
        :param threads_p:  A pointer to an underlying *notmuch_threads_t*
             structure. These are not publicly exposed, so a user
             will almost never instantiate a :class:`Threads` object
             herself. They are usually handed back as a result,
             e.g. in :meth:`Query.search_threads`.  *threads_p* must be
             valid, we will raise an :exc:`NullPointerError` if it is
             `None`.
        :type threads_p: :class:`ctypes.c_void_p`
        :param parent: The parent object
             (ie :class:`Query`) these tags are derived from. It saves
             a reference to it, so we can automatically delete the db
             object once all derived objects are dead.
        :TODO: Make the iterator work more than once and cache the tags in
               the Python object.(?)
        """
        if not threads_p:
            raise NullPointerError()

        self._threads = threads_p
        #store parent, so we keep them alive as long as self  is alive
        self._parent = parent

    def __iter__(self):
        """ Make Threads an iterator """
        return self

    _valid = nmlib.notmuch_threads_valid
    _valid.argtypes = [NotmuchThreadsP]
    _valid.restype = bool

    _move_to_next = nmlib.notmuch_threads_move_to_next
    _move_to_next.argtypes = [NotmuchThreadsP]
    _move_to_next.restype = None

    def __next__(self):
        if not self._threads:
            raise NotInitializedError()

        if not self._valid(self._threads):
            self._threads = None
            raise StopIteration

        thread = Thread(Threads._get(self._threads), self)
        self._move_to_next(self._threads)
        return thread
    next = __next__ # python2.x iterator protocol compatibility

    def __nonzero__(self):
        '''
        Implement truth value testing. If __nonzero__ is not
        implemented, the python runtime would fall back to `len(..) >
        0` thus exhausting the iterator.

        :returns: True if the wrapped iterator has at least one more object
                  left.
        '''
        return self._threads and self._valid(self._threads)

    _destroy = nmlib.notmuch_threads_destroy
    _destroy.argtypes = [NotmuchThreadsP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch Threads"""
        if self._threads:
            self._destroy(self._threads)




# -*- coding: utf-8 -*-

import sys
import os
from pathlib import Path
sys.path.append(str(Path(__file__).parent))

extensions = [ 'sphinx.ext.autodoc', 'elisp' ]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'notmuch'
copyright = u'2009-2024, Carl Worth and many others'

location = os.path.dirname(__file__)

for pathdir in ['.', '..']:
    version_file = os.path.join(location,pathdir,'version.txt')
    if os.path.exists(version_file):
        with open(version_file,'r') as infile:
            version=infile.read().replace('\n','')

# for autodoc
sys.path.insert(0, os.path.join(location, '..', 'bindings', 'python-cffi', 'build', 'stage'))

# read generated config
for pathdir in ['.', '..']:
    conf_file = os.path.join(location,pathdir,'sphinx.config')
    if os.path.exists(conf_file):
        with open(conf_file,'r') as infile:
            exec(''.join(infile.readlines()))

# The full version, including alpha/beta/rc tags.
release = version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

if tags.has('WITH_EMACS'):
    # Hacky reimplementation of include to workaround limitations of
    # sphinx-doc
    lines = ['.. include:: /../emacs/rstdoc.rsti\n\n'] # in the source tree
    for file in ('notmuch.rsti', 'notmuch-lib.rsti', 'notmuch-hello.rsti', 'notmuch-show.rsti', 'notmuch-tag.rsti', 'notmuch-tree.rsti'):
        lines.extend(open(rsti_dir+'/'+file))
    rst_epilog = ''.join(lines)
    del lines
else:
    # If we don't have emacs (or the user configured --without-emacs),
    # don't build the notmuch-emacs docs, as they need emacs to generate
    # the docstring include files
    exclude_patterns.append('notmuch-emacs.rst')

if not tags.has('WITH_PYTHON'):
    exclude_patterns.append('python-bindings.rst')

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'default'

# prevent generation of python module index
html_domain_indices=[]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Output file base name for HTML help builder.
htmlhelp_basename = 'notmuchdoc'

# Disable SmartyPants, as it mangles command lines.
# Despite the name, this actually affects manual pages as well.
html_use_smartypants = False

# See:
# - https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-manpages_url
# - https://manpages.debian.org/
manpages_url = 'https://manpages.debian.org/{page}.{section}.html'

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).

notmuch_authors = u'Carl Worth and many others'

man_make_section_directory = False

man_pages = [
    ('man1/notmuch', 'notmuch',
     u'thread-based email index, search, and tagging',
     [notmuch_authors], 1),

    ('man1/notmuch-address', 'notmuch-address',
     u'output addresses from matching messages',
     [notmuch_authors], 1),

    ('man1/notmuch-compact', 'notmuch-compact',
     u'compact the notmuch database',
     [notmuch_authors], 1),

    ('man1/notmuch-config', 'notmuch-config',
     u'access notmuch configuration file',
     [notmuch_authors], 1),

    ('man1/notmuch-count', 'notmuch-count',
     u'count messages matching the given search terms',
     [notmuch_authors], 1),

    ('man1/notmuch-dump', 'notmuch-dump',
     u'creates a plain-text dump of the tags of each message',
     [notmuch_authors], 1),

    ('man1/notmuch-emacs-mua', 'notmuch-emacs-mua',
     u'send mail with notmuch and emacs',
     [notmuch_authors], 1),

    ('man1/notmuch-git', 'notmuch-git',
     u'manage notmuch tags with git',
     [notmuch_authors], 1),

    ('man1/notmuch-git', 'nmbug',
     u'manage notmuch bugs with git',
     [notmuch_authors], 1),

    ('man5/notmuch-hooks', 'notmuch-hooks',
     u'hooks for notmuch',
     [notmuch_authors], 5),

    ('man1/notmuch-insert', 'notmuch-insert',
     u'add a message to the maildir and notmuch database',
     [notmuch_authors], 1),

    ('man1/notmuch-new', 'notmuch-new',
     u'incorporate new mail into the notmuch database',
     [notmuch_authors], 1),

    ('man7/notmuch-properties', 'notmuch-properties',
     u'notmuch message property conventions and documentation',
     [notmuch_authors], 7),

    ('man1/notmuch-reindex', 'notmuch-reindex',
     u're-index matching messages',
     [notmuch_authors], 1),

    ('man1/notmuch-reply', 'notmuch-reply',
     u'constructs a reply template for a set of messages',
     [notmuch_authors], 1),

    ('man1/notmuch-restore', 'notmuch-restore',
     u'restores the tags from the given file (see notmuch dump)',
     [notmuch_authors], 1),

    ('man1/notmuch-search', 'notmuch-search',
     u'search for messages matching the given search terms',
     [notmuch_authors], 1),

    ('man7/notmuch-search-terms', 'notmuch-search-terms',
     u'syntax for notmuch queries',
     [notmuch_authors], 7),

    ('man1/notmuch', 'notmuch-setup',
     u'getting started with notmuch',
     [notmuch_authors], 1),

    ('man7/notmuch-sexp-queries', 'notmuch-sexp-queries',
     u's-expression syntax for notmuch queries',
     [notmuch_authors], 7),

    ('man1/notmuch-show', 'notmuch-show',
     u'show messages matching the given search terms',
     [notmuch_authors], 1),

    ('man1/notmuch-tag', 'notmuch-tag',
     u'add/remove tags for all messages matching the search terms',
     [notmuch_authors], 1),
]

# If true, show URL addresses after external links.
#man_show_urls = False

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# If true, do not generate a @detailmenu in the "Top" node's menu.
texinfo_no_detailmenu = True

texinfo_documents = [
    ('notmuch-emacs', 'notmuch-emacs', u'notmuch-emacs documentation',
     notmuch_authors, 'notmuch-emacs',
     'emacs based front-end for notmuch', 'Miscellaneous'),
]

# generate texinfo list from man page list
texinfo_documents += [
    (
        x[0],				# source start file
        x[1],				# target name
        x[1] + u' documentation',	# title
        x[3][0],			# author
        x[1],				# dir menu entry
        x[2],				# description
        'Miscellaneous'			# category
    ) for x in man_pages]

def setup(app):
    import docutils.nodes
    # define nmconfig role and directive for config items.
    app.add_object_type('nmconfig','nmconfig',
                        indextemplate='pair: configuration item; %s',
                        ref_nodeclass=docutils.nodes.generated,
                        objname='config item' )



# This gdb Python script runs notmuch new and simulates killing and
# restarting notmuch new after every Xapian commit.  To simulate this
# more efficiently, this script runs notmuch new and, immediately
# after every Xapian commit, it *pauses* the running notmuch new,
# copies the entire database and maildir to a snapshot directory, and
# executes a full notmuch new on that snapshot, comparing the final
# results with the expected output.  It can then resume the paused
# notmuch new, which is still running on the original maildir, and
# repeat this process.

import gdb
import os
import glob
import shutil
import subprocess

gdb.execute('set args new')

# Make Xapian commit after every operation instead of batching
gdb.execute('set environment XAPIAN_FLUSH_THRESHOLD = 1')

maildir = os.environ['MAIL_DIR']

# Trap calls to rename, which happens just before Xapian commits
class RenameBreakpoint(gdb.Breakpoint):
    def __init__(self, *args, **kwargs):
        super(RenameBreakpoint, self).__init__(*args, **kwargs)
        self.last_inodes = {}
        self.n = 0

    def stop(self):
        xapiandir = '%s/.notmuch/xapian' % maildir
        if os.path.isfile('%s/iamchert' % xapiandir):
            # As an optimization, only consider snapshots after a
            # Xapian has really committed.  The chert backend
            # overwrites record.base? as the last step in the commit,
            # so keep an eye on their inumbers.
            inodes = {}
            for path in glob.glob('%s/record.base*' % xapiandir):
                inodes[path] = os.stat(path).st_ino
            if inodes == self.last_inodes:
                # Continue
                return False
            self.last_inodes = inodes

        # Save a backtrace in case the test does fail
        backtrace = gdb.execute('backtrace', to_string=True)
        open('backtrace.%d' % self.n, 'w').write(backtrace)

        # Snapshot the database
        shutil.rmtree('%s.snap/.notmuch' % maildir)
        shutil.copytree('%s/.notmuch' % maildir, '%s.snap/.notmuch' % maildir)
        # Restore the mtime of $MAIL_DIR.snap/
        shutil.copystat('%s/.notmuch' % maildir, '%s.snap/.notmuch' % maildir)

        # Run notmuch new to completion on the snapshot
        env = os.environ.copy()
        env.update(NOTMUCH_CONFIG=os.environ['NOTMUCH_CONFIG'] + '.snap',
                   XAPIAN_FLUSH_THRESHOLD='1000')
        subprocess.check_call(
            ['notmuch', 'new'], env=env, stdout=open('/dev/null', 'w'))
        subprocess.check_call(
            ['notmuch', 'search', '*'], env=env,
            stdout=open('search.%d' % self.n, 'w'))

        # Tell the shell how far we've gotten
        open('outcount', 'w').write(str(self.n + 1))

        # Continue
        self.n += 1
        return False
RenameBreakpoint('rename')

try:
    gdb.execute('run')
except Exception:
    import traceback
    raise SystemExit(traceback.format_exc())



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from ctypes import CDLL, Structure, POINTER
from notmuch.version import SOVERSION

#-----------------------------------------------------------------------------
#package-global instance of the notmuch library
try:
    from os import uname
    if uname()[0] == 'Darwin':
        nmlib = CDLL("libnotmuch.{0:s}.dylib".format(SOVERSION))
    else:
        nmlib = CDLL("libnotmuch.so.{0:s}".format(SOVERSION))
except:
    raise ImportError("Could not find shared 'notmuch' library.")

from .compat import Python3StringMixIn, encode_utf8 as _str

# We import these on behalf of other modules.  Silence warning about
# these symbols not being used.
Python3StringMixIn
_str

class Enum(object):
    """Provides ENUMS as "code=Enum(['a','b','c'])" where code.a=0 etc..."""
    def __init__(self, names):
        for number, name in enumerate(names):
            setattr(self, name, number)


class NotmuchDatabaseS(Structure):
    pass
NotmuchDatabaseP = POINTER(NotmuchDatabaseS)


class NotmuchQueryS(Structure):
    pass
NotmuchQueryP = POINTER(NotmuchQueryS)


class NotmuchThreadsS(Structure):
    pass
NotmuchThreadsP = POINTER(NotmuchThreadsS)


class NotmuchThreadS(Structure):
    pass
NotmuchThreadP = POINTER(NotmuchThreadS)


class NotmuchMessagesS(Structure):
    pass
NotmuchMessagesP = POINTER(NotmuchMessagesS)


class NotmuchMessageS(Structure):
    pass
NotmuchMessageP = POINTER(NotmuchMessageS)


class NotmuchMessagePropertiesS(Structure):
    pass
NotmuchMessagePropertiesP = POINTER(NotmuchMessagePropertiesS)


class NotmuchTagsS(Structure):
    pass
NotmuchTagsP = POINTER(NotmuchTagsS)


class NotmuchDirectoryS(Structure):
    pass
NotmuchDirectoryP = POINTER(NotmuchDirectoryS)


class NotmuchFilenamesS(Structure):
    pass
NotmuchFilenamesP = POINTER(NotmuchFilenamesS)


class NotmuchConfigListS(Structure):
    pass
NotmuchConfigListP = POINTER(NotmuchConfigListS)


class NotmuchIndexoptsS(Structure):
    pass
NotmuchIndexoptsP = POINTER(NotmuchIndexoptsS)



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

from ctypes import c_char_p, c_long, c_int
from .globals import (
    nmlib,
    NotmuchThreadP,
    NotmuchMessagesP,
    NotmuchTagsP,
)
from .errors import (
    NullPointerError,
    NotInitializedError,
)
from .messages import Messages
from .tag import Tags
from datetime import date

class Thread(object):
    """Represents a single message thread."""

    """notmuch_thread_get_thread_id"""
    _get_thread_id = nmlib.notmuch_thread_get_thread_id
    _get_thread_id.argtypes = [NotmuchThreadP]
    _get_thread_id.restype = c_char_p

    """notmuch_thread_get_authors"""
    _get_authors = nmlib.notmuch_thread_get_authors
    _get_authors.argtypes = [NotmuchThreadP]
    _get_authors.restype = c_char_p

    """notmuch_thread_get_subject"""
    _get_subject = nmlib.notmuch_thread_get_subject
    _get_subject.argtypes = [NotmuchThreadP]
    _get_subject.restype = c_char_p

    """notmuch_thread_get_toplevel_messages"""
    _get_toplevel_messages = nmlib.notmuch_thread_get_toplevel_messages
    _get_toplevel_messages.argtypes = [NotmuchThreadP]
    _get_toplevel_messages.restype = NotmuchMessagesP

    _get_newest_date = nmlib.notmuch_thread_get_newest_date
    _get_newest_date.argtypes = [NotmuchThreadP]
    _get_newest_date.restype = c_long

    _get_oldest_date = nmlib.notmuch_thread_get_oldest_date
    _get_oldest_date.argtypes = [NotmuchThreadP]
    _get_oldest_date.restype = c_long

    """notmuch_thread_get_tags"""
    _get_tags = nmlib.notmuch_thread_get_tags
    _get_tags.argtypes = [NotmuchThreadP]
    _get_tags.restype = NotmuchTagsP

    def __init__(self, thread_p, parent=None):
        """
        :param thread_p: A pointer to an internal notmuch_thread_t
            Structure.  These are not publicly exposed, so a user
            will almost never instantiate a :class:`Thread` object
            herself. They are usually handed back as a result,
            e.g. when iterating through :class:`Threads`. *thread_p*
            must be valid, we will raise an :exc:`NullPointerError`
            if it is `None`.

        :param parent: A 'parent' object is passed which this message is
              derived from. We save a reference to it, so we can
              automatically delete the parent object once all derived
              objects are dead.
        """
        if not thread_p:
            raise NullPointerError()
        self._thread = thread_p
        #keep reference to parent, so we keep it alive
        self._parent = parent

    def get_thread_id(self):
        """Get the thread ID of 'thread'

        The returned string belongs to 'thread' and will only be valid
        for as long as the thread is valid.

        :returns: String with a message ID
        :raises: :exc:`NotInitializedError` if the thread
                    is not initialized.
        """
        if not self._thread:
            raise NotInitializedError()
        return Thread._get_thread_id(self._thread).decode('utf-8', 'ignore')

    _get_total_messages = nmlib.notmuch_thread_get_total_messages
    _get_total_messages.argtypes = [NotmuchThreadP]
    _get_total_messages.restype = c_int

    def get_total_messages(self):
        """Get the total number of messages in 'thread'

        :returns: The number of all messages in the database
                  belonging to this thread. Contrast with
                  :meth:`get_matched_messages`.
        :raises: :exc:`NotInitializedError` if the thread
                    is not initialized.
        """
        if not self._thread:
            raise NotInitializedError()
        return self._get_total_messages(self._thread)

    def get_toplevel_messages(self):
        """Returns a :class:`Messages` iterator for the top-level messages in
           'thread'

           This iterator will not necessarily iterate over all of the messages
           in the thread. It will only iterate over the messages in the thread
           which are not replies to other messages in the thread.

        :returns: :class:`Messages`
        :raises: :exc:`NotInitializedError` if query is not initialized
        :raises: :exc:`NullPointerError` if search_messages failed
        """
        if not self._thread:
            raise NotInitializedError()

        msgs_p = Thread._get_toplevel_messages(self._thread)

        if not msgs_p:
            raise NullPointerError()

        return Messages(msgs_p, self)

    """notmuch_thread_get_messages"""
    _get_messages = nmlib.notmuch_thread_get_messages
    _get_messages.argtypes = [NotmuchThreadP]
    _get_messages.restype = NotmuchMessagesP

    def get_messages(self):
        """Returns a :class:`Messages` iterator for all messages in 'thread'

        :returns: :class:`Messages`
        :raises: :exc:`NotInitializedError` if query is not initialized
        :raises: :exc:`NullPointerError` if get_messages failed
        """
        if not self._thread:
            raise NotInitializedError()

        msgs_p = Thread._get_messages(self._thread)

        if not msgs_p:
            raise NullPointerError()

        return Messages(msgs_p, self)

    _get_matched_messages = nmlib.notmuch_thread_get_matched_messages
    _get_matched_messages.argtypes = [NotmuchThreadP]
    _get_matched_messages.restype = c_int

    def get_matched_messages(self):
        """Returns the number of messages in 'thread' that matched the query

        :returns: The number of all messages belonging to this thread that
                  matched the :class:`Query`from which this thread was created.
                  Contrast with :meth:`get_total_messages`.
        :raises: :exc:`NotInitializedError` if the thread
                    is not initialized.
        """
        if not self._thread:
            raise NotInitializedError()
        return self._get_matched_messages(self._thread)

    def get_authors(self):
        """Returns the authors of 'thread'

        The returned string is a comma-separated list of the names of the
        authors of mail messages in the query results that belong to this
        thread.

        The returned string belongs to 'thread' and will only be valid for
        as long as this Thread() is not deleted.
        """
        if not self._thread:
            raise NotInitializedError()
        authors = Thread._get_authors(self._thread)
        if not authors:
            return None
        return authors.decode('UTF-8', 'ignore')

    def get_subject(self):
        """Returns the Subject of 'thread'

        The returned string belongs to 'thread' and will only be valid for
        as long as this Thread() is not deleted.
        """
        if not self._thread:
            raise NotInitializedError()
        subject = Thread._get_subject(self._thread)
        if not subject:
            return None
        return subject.decode('UTF-8', 'ignore')

    def get_newest_date(self):
        """Returns time_t of the newest message date

        :returns: A time_t timestamp.
        :rtype: c_unit64
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._thread:
            raise NotInitializedError()
        return Thread._get_newest_date(self._thread)

    def get_oldest_date(self):
        """Returns time_t of the oldest message date

        :returns: A time_t timestamp.
        :rtype: c_unit64
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._thread:
            raise NotInitializedError()
        return Thread._get_oldest_date(self._thread)

    def get_tags(self):
        """ Returns the message tags

        In the Notmuch database, tags are stored on individual
        messages, not on threads. So the tags returned here will be all
        tags of the messages which matched the search and which belong to
        this thread.

        The :class:`Tags` object is owned by the thread and as such, will only
        be valid for as long as this :class:`Thread` is valid (e.g. until the
        query from which it derived is explicitly deleted).

        :returns: A :class:`Tags` iterator.
        :raises: :exc:`NotInitializedError` if query is not initialized
        :raises: :exc:`NullPointerError` if search_messages failed
        """
        if not self._thread:
            raise NotInitializedError()

        tags_p = Thread._get_tags(self._thread)
        if not tags_p:
            raise NullPointerError()
        return Tags(tags_p, self)

    def __unicode__(self):
        frm = "thread:%s %12s [%d/%d] %s; %s (%s)"

        return frm % (self.get_thread_id(),
                      date.fromtimestamp(self.get_newest_date()),
                      self.get_matched_messages(),
                      self.get_total_messages(),
                      self.get_authors(),
                      self.get_subject(),
                      self.get_tags(),
                     )

    _destroy = nmlib.notmuch_thread_destroy
    _destroy.argtypes = [NotmuchThreadP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch Thread"""
        if self._thread:
            self._destroy(self._thread)



#!/usr/bin/env python3

"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
"""

import os
from distutils.core import setup

# get the notmuch version number without importing the notmuch module
version_file = os.path.join(os.path.dirname(__file__),
                            'notmuch', 'version.py')
exec(compile(open(version_file).read(), version_file, 'exec'))
assert '__VERSION__' in globals(), \
    'Failed to read the notmuch binding version number'

setup(name='notmuch',
      version=__VERSION__,
      description='Python binding of the notmuch mail search and indexing library.',
      author='Sebastian Spaeth',
      author_email='Sebastian@SSpaeth.de',
      url='https://notmuchmail.org/',
      download_url='https://notmuchmail.org/releases/notmuch-%s.tar.gz' % __VERSION__,
      packages=['notmuch'],
      keywords=['library', 'email'],
      long_description='''Overview
========

The notmuch module provides an interface to the `notmuch
<https://notmuchmail.org>`_ functionality, directly interfacing with a
shared notmuch library. Notmuch provides a maildatabase that allows
for extremely quick searching and filtering of your email according to
various criteria.

The documentation for the latest notmuch release can be `viewed
online <https://notmuch.readthedocs.io/>`_.

Requirements
------------

You need to have notmuch installed (or rather libnotmuch.so.1). Also,
notmuch makes use of the ctypes library, and has only been tested with
python >= 2.5. It will not work on earlier python versions.
''',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 3',
                   'Topic :: Communications :: Email',
                   'Topic :: Software Development :: Libraries'
                   ],
      platforms='',
      license='https://www.gnu.org/licenses/gpl-3.0.txt',
     )



"""
This file is part of notmuch.

Notmuch is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Notmuch is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with notmuch.  If not, see <https://www.gnu.org/licenses/>.

Copyright 2010 Sebastian Spaeth <Sebastian@SSpaeth.de>
               Jesse Rosenthal <jrosenthal@jhu.edu>
"""


from ctypes import c_char_p, c_long, c_uint, c_int, POINTER, byref
from datetime import date
from .globals import (
    nmlib,
    Enum,
    _str,
    Python3StringMixIn,
    NotmuchTagsP,
    NotmuchMessageP,
    NotmuchMessagesP,
    NotmuchMessagePropertiesP,
    NotmuchFilenamesP,
)
from .errors import (
    STATUS,
    NotmuchError,
    NullPointerError,
    NotInitializedError,
)
from .tag import Tags
from .filenames import Filenames

import email
import sys


class Message(Python3StringMixIn):
    r"""Represents a single Email message

    Technically, this wraps the underlying *notmuch_message_t*
    structure. A user will usually not create these objects themselves
    but get them as search results.

    As it implements :meth:`__cmp__`, it is possible to compare two
    :class:`Message`\s using `if msg1 == msg2: ...`.
    """

    """notmuch_message_get_filename (notmuch_message_t *message)"""
    _get_filename = nmlib.notmuch_message_get_filename
    _get_filename.argtypes = [NotmuchMessageP]
    _get_filename.restype = c_char_p

    """return all filenames for a message"""
    _get_filenames = nmlib.notmuch_message_get_filenames
    _get_filenames.argtypes = [NotmuchMessageP]
    _get_filenames.restype = NotmuchFilenamesP

    """notmuch_message_get_flag"""
    _get_flag = nmlib.notmuch_message_get_flag
    _get_flag.argtypes = [NotmuchMessageP, c_uint]
    _get_flag.restype = bool

    """notmuch_message_set_flag"""
    _set_flag = nmlib.notmuch_message_set_flag
    _set_flag.argtypes = [NotmuchMessageP, c_uint, c_int]
    _set_flag.restype = None

    """notmuch_message_get_message_id (notmuch_message_t *message)"""
    _get_message_id = nmlib.notmuch_message_get_message_id
    _get_message_id.argtypes = [NotmuchMessageP]
    _get_message_id.restype = c_char_p

    """notmuch_message_get_thread_id"""
    _get_thread_id = nmlib.notmuch_message_get_thread_id
    _get_thread_id.argtypes = [NotmuchMessageP]
    _get_thread_id.restype = c_char_p

    """notmuch_message_get_replies"""
    _get_replies = nmlib.notmuch_message_get_replies
    _get_replies.argtypes = [NotmuchMessageP]
    _get_replies.restype = NotmuchMessagesP

    """notmuch_message_get_tags (notmuch_message_t *message)"""
    _get_tags = nmlib.notmuch_message_get_tags
    _get_tags.argtypes = [NotmuchMessageP]
    _get_tags.restype = NotmuchTagsP

    _get_date = nmlib.notmuch_message_get_date
    _get_date.argtypes = [NotmuchMessageP]
    _get_date.restype = c_long

    _get_header = nmlib.notmuch_message_get_header
    _get_header.argtypes = [NotmuchMessageP, c_char_p]
    _get_header.restype = c_char_p

    """notmuch_status_t ..._maildir_flags_to_tags (notmuch_message_t *)"""
    _tags_to_maildir_flags = nmlib.notmuch_message_tags_to_maildir_flags
    _tags_to_maildir_flags.argtypes = [NotmuchMessageP]
    _tags_to_maildir_flags.restype = c_int

    """notmuch_status_t ..._tags_to_maildir_flags (notmuch_message_t *)"""
    _maildir_flags_to_tags = nmlib.notmuch_message_maildir_flags_to_tags
    _maildir_flags_to_tags.argtypes = [NotmuchMessageP]
    _maildir_flags_to_tags.restype = c_int

    """notmuch_message_get_property"""
    _get_property = nmlib.notmuch_message_get_property
    _get_property.argtypes = [NotmuchMessageP, c_char_p, POINTER(c_char_p)]
    _get_property.restype = c_int

    """notmuch_message_get_properties"""
    _get_properties = nmlib.notmuch_message_get_properties
    _get_properties.argtypes = [NotmuchMessageP, c_char_p, c_int]
    _get_properties.restype = NotmuchMessagePropertiesP

    """notmuch_message_properties_valid"""
    _properties_valid = nmlib.notmuch_message_properties_valid
    _properties_valid.argtypes = [NotmuchMessagePropertiesP]
    _properties_valid.restype = bool

    """notmuch_message_properties_value"""
    _properties_value = nmlib.notmuch_message_properties_value
    _properties_value.argtypes = [NotmuchMessagePropertiesP]
    _properties_value.restype = c_char_p

    """notmuch_message_properties_key"""
    _properties_key = nmlib.notmuch_message_properties_key
    _properties_key.argtypes = [NotmuchMessagePropertiesP]
    _properties_key.restype = c_char_p

    """notmuch_message_properties_move_to_next"""
    _properties_move_to_next = nmlib.notmuch_message_properties_move_to_next
    _properties_move_to_next.argtypes = [NotmuchMessagePropertiesP]
    _properties_move_to_next.restype = None

    #Constants: Flags that can be set/get with set_flag
    FLAG = Enum(['MATCH'])

    def __init__(self, msg_p, parent=None):
        """
        :param msg_p: A pointer to an internal notmuch_message_t
            Structure.  If it is `None`, we will raise an
            :exc:`NullPointerError`.

        :param parent: A 'parent' object is passed which this message is
              derived from. We save a reference to it, so we can
              automatically delete the parent object once all derived
              objects are dead.
        """
        if not msg_p:
            raise NullPointerError()
        self._msg = msg_p
        #keep reference to parent, so we keep it alive
        self._parent = parent

    def get_message_id(self):
        """Returns the message ID

        :returns: String with a message ID
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()
        return Message._get_message_id(self._msg).decode('utf-8', 'ignore')

    def get_thread_id(self):
        """Returns the thread ID

        The returned string belongs to 'message' will only be valid for as
        long as the message is valid.

        This function will not return `None` since Notmuch ensures that every
        message belongs to a single thread.

        :returns: String with a thread ID
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()

        return Message._get_thread_id(self._msg).decode('utf-8', 'ignore')

    def get_replies(self):
        """Gets all direct replies to this message as :class:`Messages`
        iterator

        .. note::

            This call only makes sense if 'message' was ultimately obtained from
            a :class:`Thread` object, (such as by coming directly from the
            result of calling :meth:`Thread.get_toplevel_messages` or by any
            number of subsequent calls to :meth:`get_replies`). If this message
            was obtained through some non-thread means, (such as by a call to
            :meth:`Query.search_messages`), then this function will return
            an empty Messages iterator.

        :returns: :class:`Messages`.
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()

        msgs_p = Message._get_replies(self._msg)

        from .messages import Messages, EmptyMessagesResult

        if not msgs_p:
            return EmptyMessagesResult(self)

        return Messages(msgs_p, self)

    def get_date(self):
        """Returns time_t of the message date

        For the original textual representation of the Date header from the
        message call notmuch_message_get_header() with a header value of
        "date".

        :returns: A time_t timestamp.
        :rtype: c_unit64
        :raises: :exc:`NotInitializedError` if the message
                    is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()
        return Message._get_date(self._msg)

    def get_header(self, header):
        """Get the value of the specified header.

        The value will be read from the actual message file, not from
        the notmuch database. The header name is case insensitive.

        Returns an empty string ("") if the message does not contain a
        header line matching 'header'.

        :param header: The name of the header to be retrieved.
                       It is not case-sensitive.
        :type header: str
        :returns: The header value as string
        :raises: :exc:`NotInitializedError` if the message is not
                 initialized
        :raises: :exc:`NullPointerError` if any error occurred
        """
        if not self._msg:
            raise NotInitializedError()

        #Returns NULL if any error occurs.
        header = Message._get_header(self._msg, _str(header))
        if header == None:
            raise NullPointerError()
        return header.decode('UTF-8', 'ignore')

    def get_filename(self):
        """Returns the file path of the message file

        :returns: Absolute file path & name of the message file
        :raises: :exc:`NotInitializedError` if the message
              is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()
        return Message._get_filename(self._msg).decode('utf-8', 'ignore')

    def get_filenames(self):
        """Get all filenames for the email corresponding to 'message'

        Returns a Filenames() generator with all absolute filepaths for
        messages recorded to have the same Message-ID. These files must
        not necessarily have identical content."""
        if not self._msg:
            raise NotInitializedError()

        files_p = Message._get_filenames(self._msg)

        return Filenames(files_p, self)

    def get_flag(self, flag):
        """Checks whether a specific flag is set for this message

        The method :meth:`Query.search_threads` sets
        *Message.FLAG.MATCH* for those messages that match the
        query. This method allows us to get the value of this flag.

        :param flag: One of the :attr:`Message.FLAG` values (currently only
                     *Message.FLAG.MATCH*
        :returns: An unsigned int (0/1), indicating whether the flag is set.
        :raises: :exc:`NotInitializedError` if the message
              is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()
        return Message._get_flag(self._msg, flag)

    def set_flag(self, flag, value):
        """Sets/Unsets a specific flag for this message

        :param flag: One of the :attr:`Message.FLAG` values (currently only
                     *Message.FLAG.MATCH*
        :param value: A bool indicating whether to set or unset the flag.

        :raises: :exc:`NotInitializedError` if the message
              is not initialized.
        """
        if not self._msg:
            raise NotInitializedError()
        self._set_flag(self._msg, flag, value)

    def get_tags(self):
        """Returns the message tags

        :returns: A :class:`Tags` iterator.
        :raises: :exc:`NotInitializedError` if the message is not
                 initialized
        :raises: :exc:`NullPointerError` if any error occurred
        """
        if not self._msg:
            raise NotInitializedError()

        tags_p = Message._get_tags(self._msg)
        if not tags_p:
            raise NullPointerError()
        return Tags(tags_p, self)

    _add_tag = nmlib.notmuch_message_add_tag
    _add_tag.argtypes = [NotmuchMessageP, c_char_p]
    _add_tag.restype = c_uint

    def add_tag(self, tag, sync_maildir_flags=False):
        """Adds a tag to the given message

        Adds a tag to the current message. The maximal tag length is defined in
        the notmuch library and is currently 200 bytes.

        :param tag: String with a 'tag' to be added.

        :param sync_maildir_flags: If notmuch configuration is set to do
            this, add maildir flags corresponding to notmuch tags. See
            underlying method :meth:`tags_to_maildir_flags`. Use False
            if you want to add/remove many tags on a message without
            having to physically rename the file every time. Do note,
            that this will do nothing when a message is frozen, as tag
            changes will not be committed to the database yet.

        :returns: STATUS.SUCCESS if the tag was successfully added.
                  Raises an exception otherwise.
        :raises: :exc:`NullPointerError` if the `tag` argument is NULL
        :raises: :exc:`TagTooLongError` if the length of `tag` exceeds
                 Message.NOTMUCH_TAG_MAX)
        :raises: :exc:`ReadOnlyDatabaseError` if the database was opened
                 in read-only mode so message cannot be modified
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        status = self._add_tag(self._msg, _str(tag))

        # bail out on failure
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)

        if sync_maildir_flags:
            self.tags_to_maildir_flags()
        return STATUS.SUCCESS

    _remove_tag = nmlib.notmuch_message_remove_tag
    _remove_tag.argtypes = [NotmuchMessageP, c_char_p]
    _remove_tag.restype = c_uint

    def remove_tag(self, tag, sync_maildir_flags=False):
        """Removes a tag from the given message

        If the message has no such tag, this is a non-operation and
        will report success anyway.

        :param tag: String with a 'tag' to be removed.
        :param sync_maildir_flags: If notmuch configuration is set to do
            this, add maildir flags corresponding to notmuch tags. See
            underlying method :meth:`tags_to_maildir_flags`. Use False
            if you want to add/remove many tags on a message without
            having to physically rename the file every time. Do note,
            that this will do nothing when a message is frozen, as tag
            changes will not be committed to the database yet.

        :returns: STATUS.SUCCESS if the tag was successfully removed or if
                  the message had no such tag.
                  Raises an exception otherwise.
        :raises: :exc:`NullPointerError` if the `tag` argument is NULL
        :raises: :exc:`TagTooLongError` if the length of `tag` exceeds
                 Message.NOTMUCH_TAG_MAX)
        :raises: :exc:`ReadOnlyDatabaseError` if the database was opened
                 in read-only mode so message cannot be modified
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        status = self._remove_tag(self._msg, _str(tag))
        # bail out on error
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)

        if sync_maildir_flags:
            self.tags_to_maildir_flags()
        return STATUS.SUCCESS

    _remove_all_tags = nmlib.notmuch_message_remove_all_tags
    _remove_all_tags.argtypes = [NotmuchMessageP]
    _remove_all_tags.restype = c_uint

    def remove_all_tags(self, sync_maildir_flags=False):
        """Removes all tags from the given message.

        See :meth:`freeze` for an example showing how to safely
        replace tag values.


        :param sync_maildir_flags: If notmuch configuration is set to do
            this, add maildir flags corresponding to notmuch tags. See
            :meth:`tags_to_maildir_flags`. Use False if you want to
            add/remove many tags on a message without having to
            physically rename the file every time. Do note, that this
            will do nothing when a message is frozen, as tag changes
            will not be committed to the database yet.

        :returns: STATUS.SUCCESS if the tags were successfully removed.
                  Raises an exception otherwise.
        :raises: :exc:`ReadOnlyDatabaseError` if the database was opened
                 in read-only mode so message cannot be modified
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        status = self._remove_all_tags(self._msg)

        # bail out on error
        if status != STATUS.SUCCESS:
            raise NotmuchError(status)

        if sync_maildir_flags:
            self.tags_to_maildir_flags()
        return STATUS.SUCCESS

    _freeze = nmlib.notmuch_message_freeze
    _freeze.argtypes = [NotmuchMessageP]
    _freeze.restype = c_uint

    def get_property(self, prop):
        """ Retrieve the value for a single property key

        :param prop: The name of the property to get.
        :returns: String with the property value or None if there is no such
                  key. In the case of multiple values for the given key, the
                  first one is retrieved.
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        value = c_char_p()
        status = Message._get_property(self._msg, _str(prop), byref(value))
        if status != 0:
            raise NotmuchError(status)

        if value is None or value.value is None:
            return None
        return value.value.decode('utf-8')

    def get_properties(self, prop="", exact=False):
        """ Get the properties of the message, returning a generator of
        name, value pairs.

        The generator will yield once per value. There might be more than one
        value on each name, so the generator might yield the same name several
        times.

        :param prop: The name of the property to get. Otherwise it will return
                     the full list of properties of the message.
        :param exact: if True, require exact match with key. Otherwise
                      treat as prefix.
        :yields:  Each property values as a pair of `name, value`
        :ytype:   pairs of str
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        properties = Message._get_properties(self._msg, _str(prop), exact)
        while Message._properties_valid(properties):
            key = Message._properties_key(properties)
            value = Message._properties_value(properties)
            yield key.decode("utf-8"), value.decode("utf-8")
            Message._properties_move_to_next(properties)

    def freeze(self):
        """Freezes the current state of 'message' within the database

        This means that changes to the message state, (via :meth:`add_tag`,
        :meth:`remove_tag`, and :meth:`remove_all_tags`), will not be
        committed to the database until the message is :meth:`thaw` ed.

        Multiple calls to freeze/thaw are valid and these calls will
        "stack". That is there must be as many calls to thaw as to freeze
        before a message is actually thawed.

        The ability to do freeze/thaw allows for safe transactions to
        change tag values. For example, explicitly setting a message to
        have a given set of tags might look like this::

          msg.freeze()
          msg.remove_all_tags(False)
          for tag in new_tags:
              msg.add_tag(tag, False)
          msg.thaw()
          msg.tags_to_maildir_flags()

        With freeze/thaw used like this, the message in the database is
        guaranteed to have either the full set of original tag values, or
        the full set of new tag values, but nothing in between.

        Imagine the example above without freeze/thaw and the operation
        somehow getting interrupted. This could result in the message being
        left with no tags if the interruption happened after
        :meth:`remove_all_tags` but before :meth:`add_tag`.

        :returns: STATUS.SUCCESS if the message was successfully frozen.
                  Raises an exception otherwise.
        :raises: :exc:`ReadOnlyDatabaseError` if the database was opened
                 in read-only mode so message cannot be modified
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        status = self._freeze(self._msg)

        if STATUS.SUCCESS == status:
            # return on success
            return status

        raise NotmuchError(status)

    _thaw = nmlib.notmuch_message_thaw
    _thaw.argtypes = [NotmuchMessageP]
    _thaw.restype = c_uint

    def thaw(self):
        """Thaws the current 'message'

        Thaw the current 'message', synchronizing any changes that may have
        occurred while 'message' was frozen into the notmuch database.

        See :meth:`freeze` for an example of how to use this
        function to safely provide tag changes.

        Multiple calls to freeze/thaw are valid and these calls with
        "stack". That is there must be as many calls to thaw as to freeze
        before a message is actually thawed.

        :returns: STATUS.SUCCESS if the message was successfully frozen.
                  Raises an exception otherwise.
        :raises: :exc:`UnbalancedFreezeThawError` if an attempt was made
                 to thaw an unfrozen message. That is, there have been
                 an unbalanced number of calls to :meth:`freeze` and
                 :meth:`thaw`.
        :raises: :exc:`NotInitializedError` if message has not been
                 initialized
        """
        if not self._msg:
            raise NotInitializedError()

        status = self._thaw(self._msg)

        if STATUS.SUCCESS == status:
            # return on success
            return status

        raise NotmuchError(status)

    def is_match(self):
        """(Not implemented)"""
        return self.get_flag(Message.FLAG.MATCH)

    def tags_to_maildir_flags(self):
        """Synchronize notmuch tags to file Maildir flags

              'D' if the message has the "draft" tag
              'F' if the message has the "flagged" tag
              'P' if the message has the "passed" tag
              'R' if the message has the "replied" tag
              'S' if the message does not have the "unread" tag

        Any existing flags unmentioned in the list above will be
        preserved in the renaming.

        Also, if this filename is in a directory named "new", rename it
        to be within the neighboring directory named "cur".

        Do note that calling this method while a message is frozen might
        not work yet, as the modified tags have not been committed yet
        to the database.

        :returns: a :class:`STATUS` value. In short, you want to see
            notmuch.STATUS.SUCCESS here. See there for details."""
        if not self._msg:
            raise NotInitializedError()
        return Message._tags_to_maildir_flags(self._msg)

    def maildir_flags_to_tags(self):
        """Synchronize file Maildir flags to notmuch tags

            Flag    Action if present
            ----    -----------------
            'D'     Adds the "draft" tag to the message
            'F'     Adds the "flagged" tag to the message
            'P'     Adds the "passed" tag to the message
            'R'     Adds the "replied" tag to the message
            'S'     Removes the "unread" tag from the message

        For each flag that is not present, the opposite action
        (add/remove) is performed for the corresponding tags.  If there
        are multiple filenames associated with this message, the flag is
        considered present if it appears in one or more filenames. (That
        is, the flags from the multiple filenames are combined with the
        logical OR operator.)

        As a convenience, you can set the sync_maildir_flags parameter in
        :meth:`Database.index_file` to implicitly call this.

        :returns: a :class:`STATUS`. In short, you want to see
            notmuch.STATUS.SUCCESS here. See there for details."""
        if not self._msg:
            raise NotInitializedError()
        return Message._maildir_flags_to_tags(self._msg)

    def __repr__(self):
        """Represent a Message() object by str()"""
        return self.__str__()

    def __unicode__(self):
        format = "%s (%s) (%s)"
        return format % (self.get_header('from'),
                         self.get_tags(),
                         date.fromtimestamp(self.get_date()),
                        )

    def get_message_parts(self):
        """Output like notmuch show"""
        fp = open(self.get_filename(), 'rb')
        if sys.version_info[0] < 3:
            email_msg = email.message_from_file(fp)
        else:
            email_msg = email.message_from_binary_file(fp)
        fp.close()

        out = []
        for msg in email_msg.walk():
            if not msg.is_multipart():
                out.append(msg)
        return out

    def get_part(self, num):
        """Returns the nth message body part"""
        parts = self.get_message_parts()
        if (num <= 0 or num > len(parts)):
            return ""
        else:
            out_part = parts[(num - 1)]
            return out_part.get_payload(decode=True)

    def __hash__(self):
        """Implement hash(), so we can use Message() sets"""
        file = self.get_filename()
        if not file:
            return None
        return hash(file)

    def __cmp__(self, other):
        """Implement cmp(), so we can compare Message()s

        2 messages are considered equal if they point to the same
        Message-Id and if they point to the same file names. If 2
        Messages derive from different queries where some files have
        been added or removed, the same messages would not be considered
        equal (as they do not point to the same set of files
        any more)."""
        res = cmp(self.get_message_id(), other.get_message_id())
        if res:
            res = cmp(list(self.get_filenames()), list(other.get_filenames()))
        return res

    _destroy = nmlib.notmuch_message_destroy
    _destroy.argtypes = [NotmuchMessageP]
    _destroy.restype = None

    def __del__(self):
        """Close and free the notmuch Message"""
        if self._msg:
            self._destroy(self._msg)



import collections.abc

import notmuch2._base as base
import notmuch2._capi as capi
import notmuch2._errors as errors


__all__ = ['ImmutableTagSet', 'MutableTagSet', 'TagsIter']


class ImmutableTagSet(base.NotmuchObject, collections.abc.Set):
    """The tags associated with a message thread or whole database.

    Both a thread as well as the database expose the union of all tags
    in messages associated with them.  This exposes these as a
    :class:`collections.abc.Set` object.

    Note that due to the underlying notmuch API the performance of the
    implementation is not the same as you would expect from normal
    sets.  E.g. the :meth:`__contains__` and :meth:`__len__` are O(n)
    rather then O(1).

    Tags are internally stored as bytestrings but normally exposed as
    unicode strings using the UTF-8 encoding and the *ignore* decoder
    error handler.  However the :meth:`iter` method can be used to
    return tags as bytestrings or using a different error handler.

    Note that when doing arithmetic operations on tags, this class
    will return a plain normal set as it is no longer associated with
    the message.

    :param parent: the parent object
    :param ptr_name: the name of the attribute on the parent which will
       return the memory pointer.  This allows this object to
       access the pointer via the parent's descriptor and thus
       trigger :class:`MemoryPointer`'s memory safety.
    :param cffi_fn: the callable CFFI wrapper to retrieve the tags
       iter.  This can be one of notmuch_database_get_all_tags,
       notmuch_thread_get_tags or notmuch_message_get_tags.
    """

    def __init__(self, parent, ptr_name, cffi_fn):
        self._parent = parent
        self._ptr = lambda: getattr(parent, ptr_name)
        self._cffi_fn = cffi_fn

    def __del__(self):
        self._destroy()

    @property
    def alive(self):
        return self._parent.alive

    def _destroy(self):
        pass

    @classmethod
    def _from_iterable(cls, it):
        return set(it)

    def __iter__(self):
        """Return an iterator over the tags.

        Tags are yielded as unicode strings, decoded using the
        "ignore" error handler.

        :raises NullPointerError: If the iterator can not be created.
        """
        return self.iter(encoding='utf-8', errors='ignore')

    def iter(self, *, encoding=None, errors='strict'):
        """Aternate iterator constructor controlling string decoding.

        Tags are stored as bytes in the notmuch database, in Python
        it's easier to work with unicode strings and thus is what the
        normal iterator returns.  However this method allows you to
        specify how you would like to get the tags, defaulting to the
        bytestring representation instead of unicode strings.

        :param encoding: Which codec to use.  The default *None* does not
           decode at all and will return the unmodified bytes.
           Otherwise this is passed on to :func:`str.decode`.
        :param errors: If using a codec, this is the error handler.
           See :func:`str.decode` to which this is passed on.

        :raises NullPointerError: When things do not go as planned.
        """
        # self._cffi_fn should point either to
        # notmuch_database_get_all_tags, notmuch_thread_get_tags or
        # notmuch_message_get_tags.  nothmuch.h suggests these never
        # fail, let's handle NULL anyway.
        tags_p = self._cffi_fn(self._ptr())
        if tags_p == capi.ffi.NULL:
            raise errors.NullPointerError()
        tags = TagsIter(self, tags_p, encoding=encoding, errors=errors)
        return tags

    def __len__(self):
        return sum(1 for t in self)

    def __contains__(self, tag):
        if isinstance(tag, str):
            tag = tag.encode()
        for msg_tag in self.iter():
            if tag == msg_tag:
                return True
        else:
            return False

    def __eq__(self, other):
        return tuple(sorted(self.iter())) == tuple(sorted(other.iter()))

    def issubset(self, other):
        return self <= other

    def issuperset(self, other):
        return self >= other

    def union(self, other):
        return self | other

    def intersection(self, other):
        return self & other

    def difference(self, other):
        return self - other

    def symmetric_difference(self, other):
        return self ^ other

    def copy(self):
        return set(self)

    def __hash__(self):
        return hash(tuple(self.iter()))

    def __repr__(self):
        return '<{name} object at 0x{addr:x} tags={{{tags}}}>'.format(
            name=self.__class__.__name__,
            addr=id(self),
            tags=', '.join(repr(t) for t in self))


class MutableTagSet(ImmutableTagSet, collections.abc.MutableSet):
    """The tags associated with a message.

    This is a :class:`collections.abc.MutableSet` object which can be
    used to manipulate the tags of a message.

    Note that due to the underlying notmuch API the performance of the
    implementation is not the same as you would expect from normal
    sets.  E.g. the ``in`` operator and variants are O(n) rather then
    O(1).

    Tags are bytestrings and calling ``iter()`` will return an
    iterator yielding bytestrings.  However the :meth:`iter` method
    can be used to return tags as unicode strings, while all other
    operations accept either byestrings or unicode strings.  In case
    unicode strings are used they will be encoded using utf-8 before
    being passed to notmuch.
    """

    # Since we subclass ImmutableTagSet we inherit a __hash__.  But we
    # are mutable, setting it to None will make the Python machinery
    # recognise us as unhashable.
    __hash__ = None

    def add(self, tag):
        """Add a tag to the message.

        :param tag: The tag to add.
        :type tag: str or bytes.  A str will be encoded using UTF-8.

        :param sync_flags: Whether to sync the maildir flags with the
           new set of tags.  Leaving this as *None* respects the
           configuration set in the database, while *True* will always
           sync and *False* will never sync.
        :param sync_flags: NoneType or bool

        :raises TypeError: If the tag is not a valid type.
        :raises TagTooLongError: If the added tag exceeds the maximum
           length, see ``notmuch_cffi.NOTMUCH_TAG_MAX``.
        :raises ReadOnlyDatabaseError: If the database is opened in
           read-only mode.
        """
        if isinstance(tag, str):
            tag = tag.encode()
        if not isinstance(tag, bytes):
            raise TypeError('Not a valid type for a tag: {}'.format(type(tag)))
        ret = capi.lib.notmuch_message_add_tag(self._ptr(), tag)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def discard(self, tag):
        """Remove a tag from the message.

        :param tag: The tag to remove.
        :type tag: str of bytes.  A str will be encoded using UTF-8.
        :param sync_flags: Whether to sync the maildir flags with the
           new set of tags.  Leaving this as *None* respects the
           configuration set in the database, while *True* will always
           sync and *False* will never sync.
        :param sync_flags: NoneType or bool

        :raises TypeError: If the tag is not a valid type.
        :raises TagTooLongError: If the tag exceeds the maximum
           length, see ``notmuch_cffi.NOTMUCH_TAG_MAX``.
        :raises ReadOnlyDatabaseError: If the database is opened in
           read-only mode.
        """
        if isinstance(tag, str):
            tag = tag.encode()
        if not isinstance(tag, bytes):
            raise TypeError('Not a valid type for a tag: {}'.format(type(tag)))
        ret = capi.lib.notmuch_message_remove_tag(self._ptr(), tag)
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def clear(self):
        """Remove all tags from the message.

        :raises ReadOnlyDatabaseError: If the database is opened in
           read-only mode.
        """
        ret = capi.lib.notmuch_message_remove_all_tags(self._ptr())
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def from_maildir_flags(self):
        """Update the tags based on the state in the message's maildir flags.

        This function examines the filenames of 'message' for maildir
        flags, and adds or removes tags on 'message' as follows when
        these flags are present:

        Flag    Action if present
        ----    -----------------
        'D'     Adds the "draft" tag to the message
        'F'     Adds the "flagged" tag to the message
        'P'     Adds the "passed" tag to the message
        'R'     Adds the "replied" tag to the message
        'S'     Removes the "unread" tag from the message

        For each flag that is not present, the opposite action
        (add/remove) is performed for the corresponding tags.

        Flags are identified as trailing components of the filename
        after a sequence of ":2,".

        If there are multiple filenames associated with this message,
        the flag is considered present if it appears in one or more
        filenames. (That is, the flags from the multiple filenames are
        combined with the logical OR operator.)
        """
        ret = capi.lib.notmuch_message_maildir_flags_to_tags(self._ptr())
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)

    def to_maildir_flags(self):
        """Update the message's maildir flags based on the notmuch tags.

        If the message's filename is in a maildir directory, that is a
        directory named ``new`` or ``cur``, and has a valid maildir
        filename then the flags will be added as such:

        'D' if the message has the "draft" tag
        'F' if the message has the "flagged" tag
        'P' if the message has the "passed" tag
        'R' if the message has the "replied" tag
        'S' if the message does not have the "unread" tag

        Any existing flags unmentioned in the list above will be
        preserved in the renaming.

        Also, if this filename is in a directory named "new", rename it to
        be within the neighboring directory named "cur".

        In case there are multiple files associated with the message
        all filenames will get the same logic applied.
        """
        ret = capi.lib.notmuch_message_tags_to_maildir_flags(self._ptr())
        if ret != capi.lib.NOTMUCH_STATUS_SUCCESS:
            raise errors.NotmuchError(ret)


class TagsIter(base.NotmuchObject, collections.abc.Iterator):
    """Iterator over tags.

    This is only an iterator, not a container so calling
    :meth:`__iter__` does not return a new, replenished iterator but
    only itself.

    :param parent: The parent object to keep alive.
    :param tags_p: The CFFI pointer to the C-level tags iterator.
    :param encoding: Which codec to use.  The default *None* does not
       decode at all and will return the unmodified bytes.
       Otherwise this is passed on to :func:`str.decode`.
    :param errors: If using a codec, this is the error handler.
       See :func:`str.decode` to which this is passed on.

    :raises ObjectDestroyedError: if used after destroyed.
    """
    _tags_p = base.MemoryPointer()

    def __init__(self, parent, tags_p, *, encoding=None, errors='strict'):
        self._parent = parent
        self._tags_p = tags_p
        self._encoding = encoding
        self._errors = errors

    def __del__(self):
        self._destroy()

    @property
    def alive(self):
        if not self._parent.alive:
            return False
        try:
            self._tags_p
        except errors.ObjectDestroyedError:
            return False
        else:
            return True

    def _destroy(self):
        if self.alive:
            try:
                capi.lib.notmuch_tags_destroy(self._tags_p)
            except errors.ObjectDestroyedError:
                pass
        self._tags_p = None

    def __iter__(self):
        """Return the iterator itself.

        Note that as this is an iterator and not a container this will
        not return a new iterator.  Thus any elements already consumed
        will not be yielded by the :meth:`__next__` method anymore.
        """
        return self

    def __next__(self):
        if not capi.lib.notmuch_tags_valid(self._tags_p):
            self._destroy()
            raise StopIteration()
        tag_p = capi.lib.notmuch_tags_get(self._tags_p)
        tag = capi.ffi.string(tag_p)
        if self._encoding:
            tag = tag.decode(encoding=self._encoding, errors=self._errors)
        capi.lib.notmuch_tags_move_to_next(self._tags_p)
        return tag

    def __repr__(self):
        try:
            self._tags_p
        except errors.ObjectDestroyedError:
            return '<TagsIter (exhausted)>'
        else:
            return '<TagsIter>'
